# coding: utf-8
# create by tongshiwei on 2020-11-12

import logging

from copy import deepcopy
from mxnet import gluon


def get_trainer(net, optimizer='sgd', optimizer_params=None, lr_params=None, select=None, logger=logging,
                *args, **kwargs):
    """
    把优化器安装到网络上

    New in version 1.3.16
    """
    if lr_params:
        if "update_params" in lr_params and len(lr_params) == 1:
            pass
        else:
            from longling.ML.MxnetHelper.toolkit import get_lr_scheduler
            optimizer_params = deepcopy(optimizer_params)
            optimizer_params["lr_scheduler"] = get_lr_scheduler(logger=logger, **lr_params)

    trainer = gluon.Trainer(
        net.collect_params(select), optimizer, optimizer_params
    )
    return trainer



# coding: utf-8
# create by tongshiwei on 2020-11-19

__all__ = ["as_tmt_mx_loss", "loss_dict2tmt_mx_loss"]

import mxnet.ndarray as nd

from longling.ML.toolkit import as_tmt_loss, loss_dict2tmt_loss


def _loss2value(loss):
    return nd.mean(loss).asscalar()


# def tmt_mx_loss(loss2value=_loss2value):
#     return tmt_loss(loss2value)


def as_tmt_mx_loss(loss_obj, loss2value=_loss2value):
    return as_tmt_loss(loss_obj, loss2value)


def loss_dict2tmt_mx_loss(loss_dict, loss2value=_loss2value, exclude=None, include=None):
    return loss_dict2tmt_loss(loss_dict, loss2value, exclude, include, as_loss=as_tmt_mx_loss)



# coding: utf-8
# 2021/3/13 @ tongshiwei


__all__ = ["fit_wrapper"]


def fit_wrapper(_fit_f):
    def fit_f(net, batch_data, loss_function, trainer,
              batch_size=None, loss_monitor=None, *args, **kwargs):
        """
        Defined how each step of batch train goes

        Parameters
        ----------
        net: HybridBlock
            The network which has been initialized
            or loaded from the existed model
        batch_size: int
                The size of each batch
        batch_data: Iterable
            The batch data for train
        trainer:
            The trainer used to update the parameters of the net
        loss_function: dict of function
            The function to compute the loss for the procession
            of back propagation
        loss_monitor: LossMonitor
            Default to ``None``

        Returns
        -------

        """
        batch_size = batch_size if batch_size is not None else len(batch_data[0])

        bp_loss = _fit_f(net, batch_data, loss_function, loss_monitor, batch_size)

        trainer.zero_grad()
        bp_loss.backward()
        trainer.step()

    return fit_f



# coding: utf-8
# create by tongshiwei on 2019-9-1
__all__ = ["fit_f", "eval_f"]

import mxnet as mx
import mxnet.ndarray as nd
from mxnet import autograd
from tqdm import tqdm

from longling.ML.MxnetHelper.toolkit.ctx import split_and_load


def _fit_f(_net, _data, loss_function, loss_monitor=None):
    data, label = _data
    # todo modify the input to net
    output = _net(data)
    loss_list = []
    for name, func in loss_function.items():
        loss = func(output, label)
        loss_list.append(loss)
        # loss_value = nd.mean(loss).asscalar()
        # if loss_monitor:
        #     loss_monitor.update(name, loss_value)
    return sum(loss_list)


def eval_f(_net, test_data, ctx=mx.cpu()):
    ground_truth = []
    prediction = []

    def evaluation_function(y_true, y_pred):
        return {"evaluation_name": 0}

    for batch_data in tqdm(test_data, "evaluating"):
        ctx_data = split_and_load(
            ctx, *batch_data,
            even_split=False
        )
        for (data, label) in ctx_data:
            output = _net(data)
            pred = output
            ground_truth.extend(label.asnumpy().tolist())
            prediction.extend(pred.asnumpy().tolist())

    return evaluation_function(ground_truth, prediction)


def fit_f(net, batch_size, batch_data,
          trainer, loss_function, loss_monitor=None,
          ctx=mx.cpu(), fit_step_func=_fit_f):
    """
    Defined how each step of batch train goes

    Parameters
    ----------
    net: HybridBlock
        The network which has been initialized
        or loaded from the existed model
    batch_size: int
            The size of each batch
    batch_data: Iterable
        The batch data for train
    trainer:
        The trainer used to update the parameters of the net
    loss_function: dict of function
        The functions to compute the loss for the procession
        of back propagation
    loss_monitor: LossMonitor
        Default to ``None``
    ctx: Context or list of Context
        Defaults to ``mx.cpu()``.
    fit_step_func: function

    Returns
    -------

    """
    ctx_data = split_and_load(
        ctx, *batch_data,
        even_split=False
    )

    with autograd.record():
        for _data in ctx_data:
            bp_loss = fit_step_func(
                net, _data, loss_function, loss_monitor
            )
            bp_loss.backward()

    # todo: confirm whether the train step is equal to batch_size
    if batch_size is not None:
        trainer.step(batch_size)
    else:
        trainer.step(len(batch_data[0]))



# coding: utf-8
# create by tongshiwei on 2019-9-1
from longling import path_append

try:
    # for python module
    from .etl import transform, etl, pseudo_data_iter
    from .configuration import Configuration, ConfigurationParser
    from .sym import get_net, get_bp_loss, fit_f, eval_f
except (ImportError, SystemError):  # pragma: no cover
    # for python script
    from etl import transform, etl, pseudo_data_iter
    from configuration import Configuration, ConfigurationParser
    from sym import get_net, get_bp_loss, fit_f, eval_f

from longling.ML.PytorchHelper import set_device


def numerical_check(_net, _cfg: Configuration, train_data, test_data, dump_result=False,
                    reporthook=None, final_reporthook=None):  # pragma: no cover
    ctx = _cfg.ctx

    _net = set_device(_net, ctx)

    bp_loss_f = get_bp_loss(ctx, **_cfg.loss_params)
    loss_function = {}
    loss_function.update(bp_loss_f)

    from longling.ML.toolkit import EpochEvalFMT as Formatter
    from longling.ML.toolkit import MovingLoss
    from tqdm import tqdm

    loss_monitor = MovingLoss(loss_function)
    progress_monitor = tqdm
    if dump_result:
        from longling import config_logging
        validation_logger = config_logging(
            filename=path_append(_cfg.model_dir, "result.log"),
            logger="%s-validation" % _cfg.model_name,
            mode="w",
            log_format="%(message)s",
        )
        evaluation_formatter = Formatter(
            logger=validation_logger,
            dump_file=_cfg.validation_result_file,
        )
    else:
        evaluation_formatter = Formatter()

    # train check
    from longling.ML.PytorchHelper.toolkit.optimizer import get_trainer
    trainer = get_trainer(
        _net, optimizer=_cfg.optimizer,
        optimizer_params=_cfg.optimizer_params,
        select=_cfg.train_select
    )

    for epoch in range(_cfg.begin_epoch, _cfg.end_epoch):
        for batch_data in progress_monitor(train_data, "Epoch: %s" % epoch):
            fit_f(
                net=_net, batch_data=batch_data,
                trainer=trainer, bp_loss_f=bp_loss_f,
                loss_function=loss_function,
                loss_monitor=loss_monitor,
            )

        if epoch % 1 == 0:
            msg, data = evaluation_formatter(
                epoch=epoch,
                loss_name_value=dict(loss_monitor.items()),
                eval_name_value=eval_f(_net, test_data, ctx=ctx),
                extra_info=None,
                dump=dump_result,
            )
            print(msg)
            if reporthook is not None:
                reporthook(data)

        # optional, whether reset the loss at the end of each epoch
        loss_monitor.reset()

    if final_reporthook is not None:
        final_reporthook()


def pseudo_numerical_check(_net, _cfg):  # pragma: no cover
    datas = pseudo_data_iter(_cfg)
    numerical_check(_net, _cfg, datas, datas, dump_result=False)


def train(train_fn, test_fn, reporthook=None, final_reporthook=None, **cfg_kwargs):  # pragma: no cover
    from longling.ML.toolkit.hyper_search import prepare_hyper_search

    cfg_kwargs, reporthook, final_reporthook, tag = prepare_hyper_search(
        cfg_kwargs, Configuration, reporthook, final_reporthook, primary_key="prf:avg:f1"
    )

    _cfg = Configuration(**cfg_kwargs)
    _net = get_net(**_cfg.hyper_params)

    train_data = etl(_cfg.var2val(train_fn), params=_cfg)
    test_data = etl(_cfg.var2val(test_fn), params=_cfg)

    numerical_check(_net, _cfg, train_data, test_data, dump_result=True)


def sym_run(stage: (int, str) = "viz"):  # pragma: no cover
    if isinstance(stage, str):
        stage = {
            "pseudo": 0,
            "real": 1,
            "cli": 2,
        }[stage]

    if stage == 0:
        # ############################## Pesudo Test #################################
        cfg = Configuration(
            hyper_params={
            },
            ctx="cuda:0,1,2",
        )
        net = get_net(**cfg.hyper_params)
        pseudo_numerical_check(net, cfg)

    elif stage == 1:
        # ################################# Simple Train ###############################
        train(
            "$data_dir/train",
            "$data_dir/test",
            dataset="",
            ctx="cuda:0",
            optimizer_params={
                "lr": 0.001
            },
            hyper_params={
            },
            batch_size=16,
        )

    elif stage == 2:
        # ################################# CLI ###########################
        cfg_parser = ConfigurationParser(Configuration, commands=[train])
        cfg_kwargs = cfg_parser()
        assert "subcommand" in cfg_kwargs
        subcommand = cfg_kwargs["subcommand"]
        del cfg_kwargs["subcommand"]
        print(cfg_kwargs)
        eval("%s" % subcommand)(**cfg_kwargs)

    else:
        raise TypeError


if __name__ == '__main__':
    sym_run("pseudo")



# coding: utf-8
# 2021/8/5 @ tongshiwei


__all__ = ["pick", "tensor2list"]

import torch
from torch import Tensor


def pick(tensor: Tensor, index: Tensor, axis=-1):
    """

    Parameters
    ----------
    tensor
    index
    axis

    Returns
    -------
    >>> data = torch.tensor([[111, 112], [121, 122], [131, 132]])
    >>> index = torch.tensor([0, 1, 0])
    >>> pick(data, index)
    tensor([111, 122, 131])
    >>> data = torch.tensor([[[111, 112], [121, 122], [131, 132]], [[211, 212], [221, 222], [231, 232]]])
    >>> index = torch.tensor([[0, 1, 0], [1, 0, 0]])
    >>> pick(data, index)
    tensor([[111, 122, 131],
            [212, 221, 231]])
    """
    return torch.gather(tensor, axis, index.unsqueeze(axis)).squeeze(axis)


def tensor2list(tensor: Tensor):
    """
    Examples
    -------
    >>> tensor2list(torch.tensor([1, 2, 3]))
    [1, 2, 3]
    """
    return tensor.cpu().tolist()



# coding: utf-8
# 2020/3/3 @ tongshiwei
import yaml
from collections import OrderedDict

from .stream import as_io

__all__ = ["FoldedString", "dump_folded_yaml", "ordered_yaml_load"]


def folded_string_representer(dumper, data):
    return dumper.represent_scalar(u'tag:yaml.org,2002:str', data, style='|')


class FoldedString(str):
    pass


def represent_ordereddict(dumper, data):
    value = []

    for item_key, item_value in data.items():
        node_key = dumper.represent_data(item_key)
        node_value = dumper.represent_data(item_value)

        value.append((node_key, node_value))

    return yaml.nodes.MappingNode(u'tag:yaml.org,2002:map', value)


def dump_folded_yaml(yaml_string):
    """specially designed for arch module, should not be used in other places"""
    yaml.add_representer(FoldedString, folded_string_representer)
    yaml.add_representer(OrderedDict, represent_ordereddict)
    output = ""
    for line in yaml.dump(yaml_string, default_flow_style=False).split("\n"):
        if line.endswith("|"):
            output += (line[:-1] + ">" + "\n")
        else:
            if line:
                output += (line + "\n")
    return output


def ordered_yaml_load(stream, Loader=yaml.Loader, object_pairs_hook=OrderedDict):
    """
    Examples
    -------
    .. code-block :: python

        ordered_yaml_load("path_to_file.yaml")
        OrderedDict({"a":123})
    """

    class OrderedLoader(Loader):
        pass

    def construct_mapping(loader, node):
        loader.flatten_mapping(node)
        return object_pairs_hook(loader.construct_pairs(node))

    OrderedLoader.add_constructor(
        yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
        construct_mapping)

    with as_io(stream) as stream:
        return yaml.load(stream, OrderedLoader)



# coding: utf-8
# 2020/4/16 @ tongshiwei

import numpy as np
from sklearn.metrics import explained_variance_score, mean_absolute_error, mean_squared_error, r2_score

from longling import as_list
from .utils import POrderedDict

__all__ = ["regression_report"]


def regression_report(
        y_true, y_pred,
        metrics=None,
        sample_weight=None, multioutput="uniform_average", average_options=None,
        key_prefix="", key_suffix="",
        verbose=True,
):
    """

    Parameters
    ----------
    y_true : array-like of shape (n_samples,) or (n_samples, n_outputs)
        Ground truth (correct) target values.

    y_pred : array-like of shape (n_samples,) or (n_samples, n_outputs)
        Estimated target values.

    metrics: list of str,
        Support: evar(explained_variance), mse, rmse, mae, r2

    sample_weight : array-like of shape (n_samples,), optional
        Sample weights.

    multioutput : string in ['raw_values', 'uniform_average', 'variance_weighted'], list
        or array-like of shape (n_outputs)
        Defines aggregating of multiple output values.
        Disabled when verbose is True.
        Array-like value defines weights used to average errors.
        'raw_values' :
            Returns a full set of errors in case of multioutput input.
        'uniform_average' :
            Errors of all outputs are averaged with uniform weight.
            Alias: "macro"
        'variance_weighted':
            Only support in evar and r2.
            Scores of all outputs are averaged, weighted by the variances of each individual output.
            Alias: "vw"

    average_options: str or list
        default to macro, choices (one or many): "macro", "vw"

    key_prefix: str
    key_suffix: str
    verbose: bool
    Returns
    -------
    evar: explained variance
    mse: mean squared error
    rmse: root mean squared error
    mae: mean absolute error
    r2: r2 score

    Examples
    ---------
    >>> y_true = [[0.5, 1, 1], [-1, 1, 1], [7, -6, 1]]
    >>> y_pred = [[0, 2, 1], [-1, 2, 1], [8, -5, 1]]
    >>> regression_report(y_true, y_pred)   # doctest: +NORMALIZE_WHITESPACE
                           evar       mse      rmse  mae        r2
    0                  0.967742  0.416667  0.645497  0.5  0.965438
    1                  1.000000  1.000000  1.000000  1.0  0.908163
    2                  1.000000  0.000000  0.000000  0.0  1.000000
    uniform_average    0.989247  0.472222  0.548499  0.5  0.957867
    variance_weighted  0.983051  0.472222  0.548499  0.5  0.938257
    >>> regression_report(y_true, y_pred, verbose=False)   # doctest: +NORMALIZE_WHITESPACE
    evar: 0.989247	mse: 0.472222	rmse: 0.548499	mae: 0.500000	r2: 0.957867
    >>> regression_report(
    ...     y_true, y_pred, multioutput="variance_weighted", verbose=False
    ... )   # doctest: +NORMALIZE_WHITESPACE
    evar: 0.983051	mse: 0.472222	rmse: 0.548499	mae: 0.500000	r2: 0.938257
    >>> regression_report(y_true, y_pred, multioutput=[0.3, 0.6, 0.1], verbose=False)   # doctest: +NORMALIZE_WHITESPACE
    evar: 0.990323	mse: 0.725000	rmse: 0.793649	mae: 0.750000	r2: 0.934529
    >>> regression_report(y_true, y_pred, verbose=True)   # doctest: +NORMALIZE_WHITESPACE
                           evar       mse      rmse  mae        r2
    0                  0.967742  0.416667  0.645497  0.5  0.965438
    1                  1.000000  1.000000  1.000000  1.0  0.908163
    2                  1.000000  0.000000  0.000000  0.0  1.000000
    uniform_average    0.989247  0.472222  0.548499  0.5  0.957867
    variance_weighted  0.983051  0.472222  0.548499  0.5  0.938257
    >>> regression_report(
    ...     y_true, y_pred, verbose=True, average_options=["macro", "vw", [0.3, 0.6, 0.1]]
    ... )   # doctest: +NORMALIZE_WHITESPACE
                           evar       mse      rmse   mae        r2
    0                  0.967742  0.416667  0.645497  0.50  0.965438
    1                  1.000000  1.000000  1.000000  1.00  0.908163
    2                  1.000000  0.000000  0.000000  0.00  1.000000
    uniform_average    0.989247  0.472222  0.548499  0.50  0.957867
    variance_weighted  0.983051  0.472222  0.548499  0.50  0.938257
    weighted           0.990323  0.725000  0.793649  0.75  0.934529
    """
    legal_metrics = ["evar", "rmse", "mse", "mae", "r2"]
    if not metrics:
        metrics = legal_metrics

    _metrics = set(metrics)
    assert not _metrics - set(legal_metrics)

    y_true = np.asarray(y_true)
    y_pred = np.asarray(y_pred)

    average_options = as_list(average_options) if average_options else ["macro", "vw"]

    alias_dict = {
        "macro": "uniform_average",
        "vw": "variance_weighted",
    }

    ret = POrderedDict()

    if len(y_true.shape) > 1 and verbose:
        _ret = regression_report(
            y_true, y_pred, sample_weight=sample_weight, metrics=_metrics,
            multioutput="raw_values",
            key_prefix=key_prefix, key_suffix=key_suffix, verbose=False,
        )
        for i in range(y_true.shape[1]):
            ret[i] = {}
            for _metric in _ret.keys():
                ret[i][_metric] = _ret[_metric][i]

        for _multioutput in average_options:
            __multioutput = _multioutput if isinstance(_multioutput, list) else alias_dict.get(_multioutput,
                                                                                               _multioutput)
            _ret = regression_report(
                y_true, y_pred, metrics=_metrics, sample_weight=sample_weight,
                multioutput=__multioutput, key_prefix=key_prefix, key_suffix=key_suffix, verbose=False
            )
            _name = "weighted" if isinstance(_multioutput, list) else __multioutput
            ret[_name] = {}
            for _metric in _ret:
                ret[_name][_metric] = _ret[_metric]

    else:
        if "evar" in _metrics:
            ret[key_prefix + "evar" + key_suffix] = explained_variance_score(y_true, y_pred,
                                                                             sample_weight=sample_weight,
                                                                             multioutput=multioutput)

        if "mse" in _metrics:
            _multioutput = "uniform_average" if multioutput == "variance_weighted" else multioutput
            ret[key_prefix + "mse" + key_suffix] = mean_squared_error(y_true, y_pred, sample_weight=sample_weight,
                                                                      multioutput=_multioutput)
        if "rmse" in _metrics:
            _multioutput = "uniform_average" if multioutput == "variance_weighted" else multioutput
            ret[key_prefix + "rmse" + key_suffix] = mean_squared_error(y_true, y_pred, sample_weight=sample_weight,
                                                                       squared=False,
                                                                       multioutput=_multioutput)

        if "mae" in _metrics:
            _multioutput = "uniform_average" if multioutput == "variance_weighted" else multioutput
            ret[key_prefix + "mae" + key_suffix] = mean_absolute_error(y_true, y_pred, sample_weight=sample_weight,
                                                                       multioutput=_multioutput)

        if "r2" in metrics:
            ret[key_prefix + "r2" + key_suffix] = r2_score(y_true, y_pred, sample_weight=sample_weight,
                                                           multioutput=multioutput)

    return ret



# coding: utf-8
# 2019/9/20 @ tongshiwei

from longling.Architecture.install_proj.docs_proj import docs_proj
from longling.Architecture.install_file import pysetup, makefile, pytest, coverage
from longling.Architecture.install_file import gitignore, gitlab_ci, dockerfile, chart, travis


def py_proj(tar_dir, main_params, docs_params, docker_params=None, service_params=None, **kwargs):
    variables = {}
    variables.update(main_params)

    pysetup(tar_dir, **variables)

    makefile(tar_dir, **variables)

    pytest(tar_dir)
    coverage(tar_dir, **variables)

    gitignore("python", tar_dir)

    if docs_params:
        docs_proj(tar_dir, docs_params, __gitignore=False)

    if kwargs.get("travis") is True:
        travis(tar_dir)

    if docker_params:
        dockerfile("python-%s" % docker_params["docker_type"], tar_dir=tar_dir, **docker_params)

    if service_params:
        if service_params.get("gitlab_ci_params"):
            chart(tar_dir)
            gitlab_ci(
                atype=variables["project_type"],
                stages=service_params["gitlab_ci_params"],
                private=service_params["private"],
                tar_dir=tar_dir
            )



# coding: utf-8
# create by tongshiwei on 2019-9-1
from longling import path_append
from longling.ML import get_epoch_params_filepath
from longling.ML.MxnetHelper import save_params

try:
    # for python module
    from .sym import get_net, get_loss, fit_f, eval_f, net_viz, net_init
    from .etl import transform, etl, pseudo_data_iter
    from .configuration import Configuration, ConfigurationParser
except (ImportError, SystemError):  # pragma: no cover
    # for python script
    from sym import get_net, get_loss, fit_f, eval_f, net_viz, net_init
    from etl import transform, etl, pseudo_data_iter
    from configuration import Configuration, ConfigurationParser


def numerical_check(_net, _cfg: Configuration, train_data, test_data, dump_result=False,
                    reporthook=None, final_reporthook=None, params_save=False):  # pragma: no cover
    ctx = _cfg.ctx
    batch_size = _cfg.batch_size

    loss_function = get_loss(**_cfg.loss_params)

    from longling.ML.MxnetHelper.glue import module
    from longling.ML.toolkit import EpochEvalFMT as Formatter
    from longling.ML.toolkit import MovingLoss
    from tqdm import tqdm

    loss_monitor = MovingLoss(loss_function)
    progress_monitor = tqdm
    if dump_result:
        from longling import config_logging
        validation_logger = config_logging(
            filename=path_append(_cfg.model_dir, "result.log"),
            logger="%s-validation" % _cfg.model_name,
            mode="w",
            log_format="%(message)s",
        )
        evaluation_formatter = Formatter(
            logger=validation_logger,
            dump_file=_cfg.validation_result_file,
        )
    else:
        evaluation_formatter = Formatter()

    # train check
    trainer = module.Module.get_trainer(
        _net, optimizer=_cfg.optimizer,
        optimizer_params=_cfg.optimizer_params,
        select=_cfg.train_select,
        lr_params=_cfg.lr_params
    )

    for epoch in range(_cfg.begin_epoch, _cfg.end_epoch):
        for i, batch_data in enumerate(progress_monitor(train_data, "Epoch: %s" % epoch)):
            fit_f(
                net=_net, batch_size=batch_size, batch_data=batch_data,
                trainer=trainer,
                loss_function=loss_function,
                loss_monitor=loss_monitor,
                ctx=ctx,
            )
        if _cfg.lr_params and "update_params" in _cfg.lr_params and _cfg.end_epoch - _cfg.begin_epoch - 1 > 0:
            _cfg.logger.info("reset trainer")
            lr_params = _cfg.lr_params.pop("update_params")
            lr_update_params = dict(
                batches_per_epoch=i + 1,
                lr=_cfg.optimizer_params["learning_rate"],
                update_epoch=lr_params.get(
                    "update_epoch",
                    _cfg.end_epoch - _cfg.begin_epoch - 1
                )
            )
            lr_update_params.update(lr_params)

            trainer = module.Module.get_trainer(
                _net, optimizer=_cfg.optimizer,
                optimizer_params=_cfg.optimizer_params,
                lr_params=lr_update_params,
                select=_cfg.train_select,
                logger=_cfg.logger
            )

        if epoch % 1 == 0:
            msg, data = evaluation_formatter(
                iteration=epoch,
                loss_name_value=dict(loss_monitor.items()),
                eval_name_value=eval_f(_net, test_data, ctx=ctx),
                extra_info=None,
                dump=dump_result,
                keep={"msg", "data"}
            )
            print(msg)
            if reporthook is not None:
                reporthook(data)

        # optional
        loss_monitor.reset()

        if params_save and (epoch % _cfg.save_epoch == 0 or epoch == _cfg.end_epoch - 1):
            params_path = get_epoch_params_filepath(_cfg.model_name, epoch, _cfg.model_dir)
            _cfg.logger.info("save model params to %s, with select='%s'" % (params_path, _cfg.save_select))
            save_params(params_path, _net, select=_cfg.save_select)

    if final_reporthook is not None:
        final_reporthook()


def pseudo_numerical_check(_net, _cfg):  # pragma: no cover
    datas = pseudo_data_iter(_cfg)
    numerical_check(_net, _cfg, datas, datas, dump_result=False)


def train(train_fn, test_fn, reporthook=None, final_reporthook=None,
          primary_key="macro_avg:f1", params_save=False, **cfg_kwargs):  # pragma: no cover
    from longling.ML.toolkit.hyper_search import prepare_hyper_search

    cfg_kwargs, reporthook, final_reporthook, tag = prepare_hyper_search(
        cfg_kwargs, reporthook, final_reporthook, primary_key=primary_key, with_keys="Epoch"
    )

    _cfg = Configuration(**cfg_kwargs)
    print(_cfg)
    _net = get_net(**_cfg.hyper_params)
    net_init(_net, cfg=_cfg, **_cfg.init_params)

    train_data = etl(_cfg.var2val(train_fn), params=_cfg)
    test_data = etl(_cfg.var2val(test_fn), params=_cfg)

    numerical_check(_net, _cfg, train_data, test_data, dump_result=not tag, reporthook=reporthook,
                    final_reporthook=final_reporthook, params_save=params_save)


def sym_run(stage: (int, str) = "viz"):  # pragma: no cover
    if isinstance(stage, str):
        stage = {
            "viz": 0,
            "pseudo": 1,
            "real": 2,
            "cli": 3,
        }[stage]

    if stage <= 1:
        cfg = Configuration(
            hyper_params={}
        )
        net = get_net(**cfg.hyper_params)
        net.initialize()

        if stage == 0:
            # ############################## Net Visualization ###########################
            net_viz(net, cfg, False)
        else:
            # ############################## Pseudo Test #################################
            pseudo_numerical_check(net, cfg)

    elif stage == 2:
        # ################################# Simple Train ###############################
        import mxnet as mx
        train(
            "$data_dir/train",
            "$data_dir/test",
            ctx=mx.cpu(),
            optimizer_params={
                "learning_rate": 0.001
            },
            hyper_params_update={
            },
            batch_size=16,
            params_save=False
        )

    elif stage == 3:
        # ################################# CLI ###########################
        cfg_parser = ConfigurationParser(Configuration, commands=[train])
        cfg_kwargs = cfg_parser()
        assert "subcommand" in cfg_kwargs
        subcommand = cfg_kwargs["subcommand"]
        del cfg_kwargs["subcommand"]
        print(cfg_kwargs)
        eval("%s" % subcommand)(**cfg_kwargs)

    else:
        raise TypeError


if __name__ == '__main__':  # pragma: no cover
    sym_run("viz")



# coding: utf-8
# create by tongshiwei on 2020-11-17

from copy import deepcopy

__all__ = ["get_optimizer_cfg", "OPTIMIZERS"]

# By default, we suppose the batch_size is 16
# if the batch_size is 128, learning_rate should be 10 times default value
# e.g., for Adam, learning rate should be changed from 1e-3 to 1e-2.

OPTIMIZERS = {
    "base": (
        "Adam",
        {
            "learning_rate": 1e-3,
            "wd": 1e-4,
            "clip_gradient": 1,
        }
    ),
    "recurrent": (
        "RMSProp",
        {
            "learning_rate": 1e-4,
            "wd": 1e-4,
            "clip_gradient": 1,
        }
    ),
    "stable": (
        "sgd",
        {
            "learning_rate": 1e-2,
            "wd": 1e-4,
            "clip_gradient": 1,
            "momentum": 0.9,
        }
    ),
    "word_embedding": (
        "adagrad",
        {
            "learning_rate": 1e-2,
            "wd": 1e-4,
            "clip_gradient": 1,
        }
    )
}


def get_optimizer_cfg(name) -> tuple:
    """

    Parameters
    ----------
    name

    Returns
    -------
    optimizer_name: str
    optimizer_params: dict

    Examples
    --------
    >>> get_optimizer_cfg('base')
    ('Adam', {'learning_rate': 0.001, 'wd': 0.0001, 'clip_gradient': 1})
    """
    try:
        optimizer, optimizer_params = OPTIMIZERS[name]
    except KeyError:
        raise KeyError("Unknown optimizer name: %s, the name should be in: %s" % (name, ", ".join(OPTIMIZERS)))
    optimizer_params = deepcopy(optimizer_params)

    return optimizer, optimizer_params



# coding: utf-8
# Copyright @tongshiwei
from __future__ import absolute_import

from .configuration import Configuration, ConfigurationParser
from .etl import transform, etl
from .module import Module
from .sym import net_viz, get_loss, net_init, fit_f

__all__ = [
    "Module",
    "Configuration", "ConfigurationParser",
    "net_viz", "get_loss", "net_init", "fit_f",
    "etl", "transform"
]



# coding: utf-8
# 2020/3/3 @ tongshiwei

from longling.Architecture.install_proj.docs_proj import docs_proj
from longling.Architecture.install_file import gitignore, travis, dockerfile, gitlab_ci, chart


def web_proj(tar_dir, main_params, docs_params, docker_params=None, service_params=None, **kwargs):
    variables = {}
    variables.update(main_params)

    gitignore("web", tar_dir)

    if docs_params:
        docs_proj(tar_dir, docs_params, __gitignore=False)

    if kwargs.get("travis") is True:
        travis(tar_dir)

    if docker_params:
        dockerfile("nginx", tar_dir=tar_dir, **docker_params)

    if service_params:
        if service_params.get("gitlab_ci_params"):
            chart(tar_dir)
            gitlab_ci(
                atype=docker_params["docker_type"],
                stages=service_params["gitlab_ci_params"],
                private=service_params["private"],
                tar_dir=tar_dir,
                version_in_path=False if docker_params["docker_type"] == "vue" else True,
            )



# coding: utf-8
# 2021/3/13 @ tongshiwei


__all__ = ["as_tmt_torch_loss", "loss_dict2tmt_torch_loss"]

import torch
from longling.ML.toolkit import as_tmt_loss, loss_dict2tmt_loss


def _loss2value(loss: torch.Tensor):
    return loss.mean().item()


# def tmt_torch_loss(loss2value=_loss2value):
#     return tmt_loss(loss2value)


def as_tmt_torch_loss(loss_obj, loss2value=_loss2value):
    return as_tmt_loss(loss_obj, loss2value)


def loss_dict2tmt_torch_loss(loss_dict, loss2value=_loss2value, exclude=None, include=None):
    return loss_dict2tmt_loss(loss_dict, loss2value, exclude, include, as_loss=as_tmt_torch_loss)



# coding: utf-8
# 2020/3/4 @ tongshiwei

__all__ = ["STAGE_CANDIDATES", "OVERRIDE"]

from collections import OrderedDict

STAGE_CANDIDATES = OrderedDict({
    "build": {},
    "test": {},
    "review": {"on_stop": True, "only_master": 'n', "manual": 'n'},
    "docs": {"stage_image_port": '80', "only_master": 'y', "manual": 'y'},
    "production": {"only_master": 'y', "manual": 'y'},
})

OVERRIDE = None



# coding: utf-8
# create by tongshiwei on 2019-9-1

__all__ = ["get_net", "get_bp_loss"]

# todo: define your network symbol and back propagation loss function

import torch
from torch import nn

from longling.ML.PytorchHelper.toolkit.parallel import set_device


def get_net(**kwargs):
    return NetName(**kwargs)


def get_bp_loss(ctx, **kwargs):
    return {"L2Loss": set_device(ctx, torch.nn.MSELoss(**kwargs))}


class NetName(nn.Module):
    def __init__(self, **kwargs):
        super(NetName, self).__init__()

        with self.name_scope():
            pass

    def forward(self, *args, **kwargs):
        pass



# coding: utf-8
# 2020/4/16 @ tongshiwei

import functools
from pathlib import PurePath
import json
import math
import numpy as np
from longling import loading as _loading, as_out_io, PATH_TYPE, json_load, close_io
from longling import AddPrinter, type_from_name, config_logging, as_list
from sklearn.utils.validation import check_random_state
from sklearn.model_selection import KFold
from longling import concurrent_pool

_logger = config_logging(logger="dataset", console_log_level="info")
loading = functools.partial(_loading, src_type="text")


def _get_s_points(length: int, ratio: list) -> list:
    """
    Returns
    -------
    s_points: list of int

    Examples
    --------
    >>> _get_s_points(10, [0.1, 0.9])
    [0, 1, 10]
    """
    cum_ratios = np.cumsum([0.0] + ratio)
    s_points = [
        math.ceil(length * ratio) for ratio in cum_ratios
    ]
    return s_points


def _get_s_indices(indices: list, ratio: list) -> list:
    """
    Returns
    -------
    s_indices: list of set of int

    Examples
    --------
    >>> _get_s_indices(list(range(10)), ratio=[0.1, 0.9])
    [{0}, {1, 2, 3, 4, 5, 6, 7, 8, 9}]
    """
    s_points = _get_s_points(len(indices), ratio)
    separator_indices = [
        set(indices[s_points[i]: s_points[i + 1]])
        for i in range(len(s_points) - 1)
    ]
    return separator_indices


def shuffle_indices(indices, random_state=None) -> list:
    """
    Examples
    --------
    >>> shuffle_indices(list(range(10)), 10)
    [8, 2, 5, 6, 3, 1, 0, 7, 4, 9]
    """
    return check_random_state(random_state).permutation(indices).tolist()


def _ratio_list(ratio: (str, list)) -> list:
    """
    Examples
    --------
    >>> _ratio_list("2:1:1")
    [0.5, 0.25, 0.25]
    >>> _ratio_list("0.5:0.25:0.25")
    [0.5, 0.25, 0.25]
    >>> _ratio_list([2, 1, 1])
    [0.5, 0.25, 0.25]
    >>> _ratio_list([0.5, 0.25, 0.25])
    [0.5, 0.25, 0.25]
    """
    if isinstance(ratio, str):
        ratio = list(map(float, ratio.split(":")))

    return (np.asarray(ratio) / np.sum(ratio)).tolist()


def get_s_indices(*files, ratio: (str, list) = None, index_file=None,
                  s_indices: (PATH_TYPE, list) = None, shuffle=True, random_state=None):
    if s_indices is not None:
        if isinstance(s_indices, PATH_TYPE):
            s_indices = [set(_s_indices) for _s_indices in json_load(s_indices)]

    elif ratio is not None:
        ratio = _ratio_list(ratio)

        src = zip(*[loading(_file) for _file in files])
        indices = [i for i, _ in enumerate(src)]

        if shuffle:
            indices = shuffle_indices(indices, random_state)

        s_indices = _get_s_indices(indices, ratio)
    else:
        raise ValueError("ratio or s_indices should be specified")

    if index_file is not None:
        with as_out_io(index_file) as wf:
            json.dump([list(_s_indices) for _s_indices in s_indices], wf)

    return s_indices


def __dump_file(src, s_indices: list, tars: list, logger=_logger):
    """

    Parameters
    ----------
    s_indices: list of set of int
    src: PATH_TYPE
    tars: list of AddPrinter
    logger

    """
    logger.debug("split and dump %s: start" % src)
    for i, elem in enumerate(loading(src)):
        for j, values in enumerate(s_indices):
            if i in values:
                tars[j].add(elem)
                break
        else:  # pragma: no cover
            raise ValueError()
    logger.debug("split and dump %s: end" % src)


def _dump_file(srcs, s_indices: list, tars_list: list, logger=_logger):
    with concurrent_pool("t", 5) as pool:
        for src, tars in zip(srcs, tars_list):
            pool.submit(__dump_file, src, s_indices, tars, logger)


def to_add_printer(targets_list):
    return [
        [
            target if isinstance(target, AddPrinter) else AddPrinter(target, end='')
            for target in targets
        ] for targets in targets_list
    ]


def _target_names(*files, target_names=None, suffix, prefix=""):
    """

    Examples
    --------
    >>> files = ["x.txt"]
    >>> _target_names(*files, suffix=[".train", ".test"])
    [['x.train.txt', 'x.test.txt']]
    >>> _target_names(*files, suffix=[".train", ".test"], prefix="data/")
    [['data/x.train.txt', 'data/x.test.txt']]
    >>> _target_names(*files, suffix=[".train", ".test"], target_names=[["train.txt", "test.txt"]])
    [['train.txt', 'test.txt']]
    """
    if target_names is None:
        if not prefix:
            return [
                [
                    "%s%s" % (PurePath(_file).with_suffix(_suffix), type_from_name(_file))
                    for _suffix in suffix
                ] for _file in files
            ]
        else:
            return [
                [
                    "%s%s%s" % (prefix, PurePath(_file).with_suffix(_suffix).name, type_from_name(_file))
                    for _suffix in suffix
                ] for _file in files
            ]
    else:
        return as_list(target_names)


def _target_files(*files, target_names=None, suffix, prefix=""):
    target_names = _target_names(*files, target_names=target_names, suffix=suffix, prefix=prefix)
    return to_add_printer(target_names)


def _get_target_file_names(printers):
    return [printer.name for printer in printers]


def _src2tar_tips(srcs, tars, tips="", logger=_logger):
    for src, tar in zip(srcs, tars):
        logger.info("%s%s -> %s" % (tips, src, ",".join(_get_target_file_names(tar))))


def file_split(*files, ratio: (str, list) = None, target_files: list, index_file=None,
               s_indices: (PATH_TYPE, list) = None, shuffle=True, random_state=None, logger=_logger):
    """

    Parameters
    ----------
    files: list of PATH_TYPE
    ratio
    target_files: list of AddPrinter or list of list of AddPrinter
    index_file
    s_indices: PATH_TYPE or list of set of int

    Returns
    -------

    """
    s_indices = get_s_indices(*files, ratio=ratio, index_file=index_file, s_indices=s_indices, shuffle=shuffle,
                              random_state=random_state)

    _dump_file(files, s_indices, target_files, logger=logger)


def train_test(*files, train_size: (float, int) = 0.8, test_size: (float, int, None) = None, ratio=None,
               random_state=None, shuffle=True, target_names=None,
               suffix: list = None, prefix="", logger=_logger, **kwargs):
    """

    Parameters
    ----------
    files
    train_size: float, int, or None, (default=0.8)
        Represent the proportion of the dataset to include in the train split.
    test_size: float, int, or None
        Represent the proportion of the dataset to include in the train split.
    random_state: int, RandomState instance or None, optional (default=None)
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used by np.random.
    shuffle: boolean, optional (default=True)
        Whether or not to shuffle the data before splitting. If shuffle=False then stratify must be None
    target_names: list of PATH_TYPE
    suffix: list
    kwargs
    """
    logger.info("train_test start")
    suffix = [".train", ".test"] if suffix is None else suffix

    assert len(suffix) == 2

    target_files = _target_files(*files, target_names=target_names, suffix=suffix, prefix=prefix)

    if ratio is None:
        ratio = "%s:" % train_size

        if test_size is None:
            assert 0 <= train_size <= 1
            test_size = "%s" % (1 - train_size)
        else:
            test_size = "%s" % test_size

        ratio += test_size

    _src2tar_tips(files, target_files, "train_test: ", logger=logger)
    file_split(
        *files,
        ratio=ratio,
        target_files=target_files,
        shuffle=shuffle,
        random_state=random_state,
        **kwargs
    )
    close_io([[printer for printer in printers] for printers in target_files])
    logger.info("train_test end")


def train_valid_test(*files,
                     train_size: (float, int) = 0.8, valid_size: (float, int) = 0.1,
                     test_size: (float, int, None) = None, ratio=None,
                     random_state=None,
                     shuffle=True, target_names=None,
                     suffix: list = None, logger=_logger, prefix="", **kwargs):
    """

    Parameters
    ----------
    files
    train_size: float, int, or None, (default=0.8)
        Represent the proportion of the dataset to include in the train split.
    valid_size: float, int, or None, (default=0.1)
        Represent the proportion of the dataset to include in the valid split.
    test_size: float, int, or None
        Represent the proportion of the dataset to include in the test split.
    random_state: int, RandomState instance or None, optional (default=None)
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used by np.random.
    shuffle: boolean, optional (default=True)
        Whether or not to shuffle the data before splitting. If shuffle=False then stratify must be None
    target_names
    suffix: list
    kwargs
    """
    logger.info("train_valid_test start")
    suffix = [".train", ".valid", ".test"] if suffix is None else suffix

    assert len(suffix) == 3

    target_files = _target_files(*files, target_names=target_names, suffix=suffix, prefix=prefix)

    if ratio is None:
        ratio = "%s:%s:" % (train_size, valid_size)

        if test_size is None:
            assert 0 <= train_size + valid_size <= 1
            test_size = "%s" % (1 - train_size - valid_size)
        else:
            test_size = "%s" % test_size

        ratio += test_size

    _src2tar_tips(files, target_files, "train_valid_test: ", logger=logger)
    file_split(
        *files,
        ratio=ratio,
        target_files=target_files,
        shuffle=shuffle,
        random_state=random_state,
        **kwargs
    )
    close_io([[printer for printer in printers] for printers in target_files])
    logger.info("train_valid_test end")


def _kfold(*files, idx, indices, suffix, logger=_logger, prefix=""):
    logger.info("kfold %s start" % idx)
    s_indices = [set(_indices.tolist()) for _indices in indices]
    target_files = _target_files(
        *files, target_names=None, suffix=[".%s" % idx + _suffix for _suffix in suffix], prefix=prefix
    )
    _src2tar_tips(files, target_files, "kfold %s: " % idx, logger=logger)
    file_split(*files, s_indices=s_indices, target_files=target_files, logger=logger)
    close_io([[printer for printer in printers] for printers in target_files])
    logger.info("kfold %s end" % idx)


def kfold(*files, n_splits=5, shuffle=False, random_state=None, suffix=None, logger=_logger, prefix=""):
    splitter = KFold(n_splits=n_splits, shuffle=shuffle, random_state=random_state)
    suffix = [".train", ".test"] if suffix is None else suffix

    with concurrent_pool("t") as pool:
        for i, indices in enumerate(splitter.split(range(len([_ for _ in loading(files[0])])))):
            pool.submit(
                _kfold,
                *files,
                idx=i,
                indices=indices,
                suffix=suffix,
                logger=logger,
                prefix=prefix
            )



# coding: utf-8
# 2020/1/13 @ tongshiwei
import json
import os
import warnings
import functools
import queue
import threading
import multiprocessing as mp
from tqdm import tqdm

from longling import wf_open, loading

__all__ = ["BaseIter", "MemoryIter", "LoopIter", "AsyncLoopIter", "AsyncIter", "CacheAsyncLoopIter", "iterwrap"]


class Register(dict):  # pragma: no cover
    def add(self, cls):
        if cls.__name__ in self:
            warnings.warn("key %s has already existed, which will be overridden" % cls.__name__)
        self[cls.__name__] = cls
        return cls


register = Register()


@register.add
class BaseIter(object):
    """
    迭代器

    Notes
    ------

    * 如果 src 是一个迭代器实例，那么在一轮迭代之后，迭代器里的内容就被迭代完了，将无法重启。
    * 如果想使得迭代器可以一直被循环迭代，那么 src 应当是迭代器实例的生成函数, 同时在每次循环结束后，调用reset()
    * 如果 src 没有 __length__，那么在第一次迭代结束前，无法对 BaseIter 的实例调用 len() 函数

    Examples
    --------

    .. code-block:: python

        # 单次迭代后穷尽内容
        with open("demo.txt") as f:
            bi = BaseIter(f)
            for line in bi:
                pass

        # 可多次迭代
        def open_file():
            with open("demo.txt") as f:
                for line in f:
                    yield line

        bi = BaseIter(open_file)
        for _ in range(5):
            for line in bi:
                pass
            bi.reset()


        # 简化的可多次迭代的写法
        @BaseIter.wrap
        def open_file():
            with open("demo.txt") as f:
                for line in f:
                    yield line

        bi = open_file()
        for _ in range(5):
            for line in bi:
                pass
            bi.reset()
    """

    def __init__(self, src, length=None, *args, **kwargs):
        self._reset = src
        self._data = None
        self._length = length
        self._count = 0
        self.init()

    def __next__(self):
        try:
            self._count += 1
            return next(self._data)
        except StopIteration:
            self._count -= 1
            self._set_length()
            raise StopIteration

    def __iter__(self):
        return self

    def __len__(self):
        if self._length is None:
            raise TypeError("length is unknown")
        return self._length

    def init(self):
        self.reset()

    def reset(self):
        _data = self._reset() if callable(self._reset) else self._reset
        try:
            self._length = len(_data)
        except TypeError:
            self._set_length()
        self._data = iter(_data)

    def _set_length(self):
        if self._length is None and self._count > 0:
            self._length = self._count

    @classmethod
    def wrap(cls, f):
        @functools.wraps(f)
        def _f(*args, **kwargs):
            return cls(functools.partial(f, *args, **kwargs))

        return _f


@register.add
class MemoryIter(BaseIter):
    """
    内存迭代器

    会将所有迭代器内容装载入内存
    """

    def __init__(self, src, length=None, prefetch=False, *args, **kwargs):
        self._memory_data = []
        self._in_memory = True
        super(MemoryIter, self).__init__(src, length)
        self.prefetch = prefetch
        if self.prefetch:
            self._prefetch()

    def _prefetch(self):
        for _data in tqdm(self._data, "prefetching data"):
            self._count += 1
            self._memory_data.append(_data)
        self._set_length()
        self._in_memory = False
        self._data = iter(self._memory_data)

    def __next__(self):
        try:
            self._count += 1
            elem = next(self._data)
            if self._in_memory is True:
                self._memory_data.append(elem)
            return elem
        except StopIteration:
            self._count -= 1
            self.reset()
            self._in_memory = False
            raise StopIteration

    def reset(self):
        if not self._memory_data:
            super(MemoryIter, self).reset()
        else:
            self._data = iter(self._memory_data)
            self._set_length()


@register.add
class LoopIter(BaseIter):
    """
    循环迭代器

    每次迭代后会进行自动的 reset() 操作
    """

    def __init__(self, src, length=None, *args, **kwargs):
        super(LoopIter, self).__init__(src, length)

    def __next__(self):
        try:
            self._count += 1
            return next(self._data)
        except StopIteration:
            self._count -= 1
            self.reset()
            raise StopIteration


def produce(data, produce_queue):
    _stop = False
    try:
        for _data in data:
            produce_queue.put(_data)
        raise StopIteration

    except StopIteration as e:
        if not _stop:
            _stop = True
            produce_queue.put(e)

    except Exception as e:  # pragma: no cover
        if not _stop:
            _stop = True
            produce_queue.put(e)


@register.add
class AsyncLoopIter(LoopIter):
    """
    异步循环迭代器,适用于加载文件

    数据的读入和数据的使用迭代是异步的。reset() 之后会进行数据预取
    """

    def __init__(self, src, tank_size=8, timeout=None, level="t"):
        self.thread = None
        self._size = tank_size
        self.mode = self._mode_map(level)
        if self.mode == "t":
            self.queue_cls = queue.Queue
            self.thread_cls = threading.Thread
        elif self.mode == "p":
            self.queue_cls = mp.Queue
            self.thread_cls = mp.Process
        else:  # pragma: no cover
            raise TypeError("unknown mode: %s" % self.mode)
        self.queue = self.queue_cls(self._size)
        self._timeout = timeout
        super(AsyncLoopIter, self).__init__(src)

    @classmethod
    def _mode_map(cls, mode):
        map_dict = {
            "p": "p", "processing": "p", "multiprocessing": "p",
            "t": "t", "thread": "t", "threading": "t",
        }
        return map_dict[mode]

    def reset(self):
        super(AsyncLoopIter, self).reset()
        if self.thread is not None:
            self.thread.join()

        self.thread = self.thread_cls(
            target=produce,
            kwargs=dict(data=self._data, produce_queue=self.queue),
            daemon=True
        )
        self.thread.start()

    def __next__(self):
        if self.queue is not None:
            item = self.queue.get()
        else:  # pragma: no cover
            raise StopIteration
        if isinstance(item, Exception):
            if isinstance(item, StopIteration):
                self.reset()
                raise StopIteration
            else:  # pragma: no cover
                raise item
        else:
            self._count += 1
            return item


@register.add
class AsyncIter(AsyncLoopIter):
    """
    异步装载迭代器

    不会进行自动 reset()
    """

    def init(self):
        super(AsyncIter, self).reset()

    def reset(self):
        self.queue = queue.Queue(self._size)
        super(AsyncIter, self).reset()

    def __next__(self):
        if self.queue is not None:
            item = self.queue.get()
        else:
            raise StopIteration
        if isinstance(item, Exception):
            if isinstance(item, StopIteration):
                self.thread = None
                self.queue = None
                self._set_length()
                raise StopIteration
            else:  # pragma: no cover
                raise item
        else:
            self._count += 1
            return item


@register.add
class CacheAsyncLoopIter(AsyncLoopIter):
    """
    带缓冲池的异步迭代器，适用于带预处理的文件

    自动 reset(),
    同时针对 src 为 function 时可能存在的复杂预处理（即异步加载取数据操作比迭代输出数据操作时间长很多），
    将异步加载中处理的预处理数据放到指定的缓冲文件中
    """

    def __init__(self, src, cache_file, rerun=True, tank_size=8, timeout=None, level="t"):
        self.cache_file = cache_file
        self.cache_queue = None
        self.cache_thread = None

        if os.path.exists(self.cache_file) and rerun is False:
            # 从已有数据中进行装载
            def src():
                return loading(self.cache_file, "jsonl")

            self._cache_stop = True
        else:
            # 重新生成数据
            self.cache_queue = queue.Queue(tank_size)
            self.cache_thread = threading.Thread(target=self.cached, daemon=False)
            self.cache_thread.start()
            self._cache_stop = False
        super(CacheAsyncLoopIter, self).__init__(src, tank_size, timeout, level)

    def init(self):
        super(CacheAsyncLoopIter, self).reset()

    def reset(self):
        self._reset = lambda: loading(self.cache_file, "jsonl")
        if self.cache_thread is not None:
            self.cache_thread.join()
            self.cache_thread = None
            self.cache_queue = None
        super(CacheAsyncLoopIter, self).reset()

    def __next__(self):
        item = self.queue.get()
        if self.cache_queue is not None:
            self.cache_queue.put(item)
        if isinstance(item, Exception):
            if isinstance(item, StopIteration):
                self.reset()
                raise StopIteration
            else:  # pragma: no cover
                raise item
        else:
            self._count += 1
            return item

    def cached(self):
        assert self.cache_queue is not None

        with wf_open(self.cache_file, mode="w") as wf:
            while True:
                data = self.cache_queue.get()
                if isinstance(data, StopIteration):
                    break
                print(json.dumps(data), file=wf)


def iterwrap(itertype: str = "AsyncLoopIter", *args, **kwargs):
    """
    迭代器装饰器，适用于希望能重复使用一个迭代器的情况，能将迭代器生成函数转换为可以重复使用的函数。
    默认使用 AsyncLoopIter。

    Examples
    --------

    .. code-block:: python

        @iterwrap()
        def open_file():
            with open("demo.txt") as f:
                for line in f:
                    yield line

        data = open_file()
        for _ in range(5):
            for line in data:
                pass
    """
    if itertype not in register:
        raise TypeError("itertype %s is unknown, the available type are %s" % (itertype, ", ".join(register)))

    def _f1(f):
        @functools.wraps(f)
        def _f2(*fargs, **fkwargs):
            return register[itertype](functools.partial(f, *fargs, **fkwargs), *args, **kwargs)

        return _f2

    return _f1



# coding: utf-8
# 2021/2/10 @ tongshiwei

from longling import path_append
from longling.ML.toolkit.hyper_search import prepare_hyper_search
from longling.ML.toolkit import EpochEvalFMT as Formatter
from longling.ML.toolkit import MovingLoss
from longling.ML.toolkit import ConsoleProgressMonitor
from tqdm import tqdm
from longling.ML import get_epoch_params_filepath
from longling.ML.const import RESULT_LOG


def train(
        net, cfg, loss_function, trainer,
        train_data, test_data=None,
        params_save=False, dump_result=False, progress_monitor=None,
        *,
        fit_f, eval_f=None, net_init=None, get_net=None, get_loss=None, get_trainer=None, save_params=None,
        enable_hyper_search=False, reporthook=None, final_reporthook=None, primary_key=None,
        eval_epoch=1, loss_dict2tmt_loss=None, epoch_lr_scheduler=None, batch_lr_scheduler=None, loss_as_dict=False,
        verbose=None, dump_cfg=None, **cfg_kwargs):
    if enable_hyper_search:
        assert get_net is not None
        cfg_kwargs, reporthook, final_reporthook, tag = prepare_hyper_search(
            cfg_kwargs, reporthook, final_reporthook, primary_key=primary_key, with_keys="Epoch", dump=params_save
        )
        dump_result = tag
        verbose = tag if verbose is None else verbose
        cfg.update(**cfg_kwargs)
        print("hyper search enabled")
        print(cfg)

    verbose = True if verbose is None else verbose
    dump_cfg = dump_cfg if dump_cfg is not None else params_save
    if dump_cfg:
        cfg.dump()

    net = net if get_net is None else get_net(**cfg.hyper_params)

    if net_init is not None:
        net_init(net, cfg=cfg, initializer_kwargs=cfg.init_params)

    train_ctx = cfg.ctx if cfg.train_ctx is None else cfg.train_ctx
    eval_ctx = cfg.ctx if cfg.eval_ctx is None else cfg.eval_ctx
    batch_size = cfg.batch_size

    loss_function = get_loss(**cfg.loss_params) if get_loss is not None else loss_function

    if isinstance(loss_function, dict):
        _loss_function = loss_function
    else:
        if hasattr(loss_function, "__name__"):
            loss_name = loss_function.__name__
        elif hasattr(loss_function, "__class__"):
            loss_name = loss_function.__class__.__name__
        else:  # pragma: no cover
            loss_name = "loss"
        loss_function = {loss_name: loss_function}
        if loss_dict2tmt_loss is not None:
            loss_function = loss_dict2tmt_loss(loss_function)
        _loss_function = list(loss_function.values())[0] if loss_as_dict is False else loss_function

    loss_monitor = MovingLoss(loss_function)

    if progress_monitor is None and loss_dict2tmt_loss is not None:
        progress_monitor = ConsoleProgressMonitor(
            indexes={
                "Loss": [name for name in loss_function]
            },
            values={
                "Loss": loss_monitor.losses
            },
            player_type="epoch",
            total_epoch=cfg.end_epoch - 1,
            silent=not verbose
        )
    elif progress_monitor is None or progress_monitor == "tqdm":
        def progress_monitor(x, e):
            return tqdm(x, "Epoch: %s" % e, disable=not verbose)

    if dump_result:
        from longling import config_logging
        validation_logger = config_logging(
            filename=path_append(cfg.model_dir, cfg.get("result_log", RESULT_LOG)),
            logger="%s-validation" % cfg.model_name,
            mode="w",
            log_format="%(message)s",
        )
        evaluation_formatter = Formatter(
            logger=validation_logger,
            dump_file=cfg.validation_result_file,
        )
    else:
        evaluation_formatter = Formatter()

    # train check
    if get_trainer is not None:
        trainer = get_trainer(
            net, optimizer=cfg.optimizer,
            optimizer_params=cfg.optimizer_params,
            select=cfg.train_select,
            lr_params=cfg.lr_params
        )
        if batch_lr_scheduler is True:
            trainer, batch_lr_scheduler = trainer
        elif epoch_lr_scheduler is True:
            trainer, epoch_lr_scheduler = trainer

    for epoch in range(cfg.begin_epoch, cfg.end_epoch):
        for i, batch_data in enumerate(progress_monitor(train_data, epoch)):
            fit_f(
                net, batch_size=batch_size, batch_data=batch_data,
                trainer=trainer,
                loss_function=_loss_function,
                loss_monitor=loss_monitor,
                ctx=train_ctx,
            )
            if batch_lr_scheduler is not None:
                batch_lr_scheduler.step()

        if cfg.lr_params and "update_params" in cfg.lr_params and cfg.end_epoch - cfg.begin_epoch - 1 > 0:
            cfg.logger.info("reset trainer")
            lr_params = cfg.lr_params.pop("update_params")
            lr_update_params = dict(
                batches_per_epoch=i + 1,
                lr=cfg.optimizer_params["learning_rate"],
                update_epoch=lr_params.get(
                    "update_epoch",
                    cfg.end_epoch - cfg.begin_epoch - 1
                )
            )
            lr_update_params.update(lr_params)

            assert get_trainer is not None
            trainer = get_trainer(
                net, optimizer=cfg.optimizer,
                optimizer_params=cfg.optimizer_params,
                lr_params=lr_update_params,
                select=cfg.train_select,
                logger=cfg.logger
            )

        if test_data is not None and epoch % eval_epoch == 0:
            msg, data = evaluation_formatter(
                iteration=epoch,
                loss_name_value=dict(loss_monitor.items()),
                eval_name_value=eval_f(net, test_data, ctx=eval_ctx, verbose=verbose, **cfg.get("eval_params", {})),
                extra_info=None,
                dump=dump_result,
                keep={"msg", "data"}
            )
            print(msg)
            if reporthook is not None:
                reporthook(data)

        # optional
        loss_monitor.reset()

        if params_save and (epoch % cfg.save_epoch == 0 or epoch == cfg.end_epoch - 1):
            assert save_params is not None
            params_path = get_epoch_params_filepath(cfg.model_name, epoch, cfg.model_dir)
            cfg.logger.info("save model params to %s, with select='%s'" % (params_path, cfg.save_select))
            save_params(params_path, net, select=cfg.save_select)

        if epoch_lr_scheduler is not None:
            epoch_lr_scheduler.step()

    if final_reporthook is not None:
        final_reporthook()



# coding: utf-8
# create by tongshiwei on 2019/4/8

__all__ = ["AttrDict", "nested_update"]


class AttrDict(dict):
    """
    Example
    -------
    >>> ad = AttrDict({'first_name': 'Eduardo'}, last_name='Pool', age=24, sports=['Soccer'])
    >>> ad
    {'first_name': 'Eduardo', 'last_name': 'Pool', 'age': 24, 'sports': ['Soccer']}
    >>> ad.first_name
    'Eduardo'
    >>> ad.age
    24
    >>> ad.age = 16
    >>> ad.age
    16
    >>> ad["age"] = 20
    >>> ad["age"]
    20
    """

    def __init__(self, *args, **kwargs):
        super(AttrDict, self).__init__(*args, **kwargs)
        for arg in args:
            if isinstance(arg, dict):
                for k, v in arg.items():
                    self[k] = v

        if kwargs:
            for k, v in kwargs.items():
                self[k] = v

    def __getattr__(self, attr):
        try:
            return self[attr]
        except KeyError as e:
            raise AttributeError(str(e))

    def __setattr__(self, key, value):
        self.__setitem__(key, value)

    def __delattr__(self, item):
        try:
            self.__delitem__(item)
        except KeyError as e:
            raise AttributeError(str(e))

    def __setitem__(self, key, value):
        super(AttrDict, self).__setitem__(key, value)
        self.__dict__.update({key: value})

    def __delitem__(self, key):
        super(AttrDict, self).__delitem__(key)
        del self.__dict__[key]


def nested_update(src: dict, update: dict):
    """
    Examples
    -------
    >>> nested_update({"a": {"x": 1}}, {"a": {"y": 2}})
    {'a': {'x': 1, 'y': 2}}
    >>> nested_update({"a": {"x": 1}}, {"a": {"x": 2}})
    {'a': {'x': 2}}
    >>> nested_update({"a": {"x": 1}}, {"b": {"y": 2}})
    {'a': {'x': 1}, 'b': {'y': 2}}
    >>> nested_update({"a": {"x": 1}}, {"a": 2})
    {'a': 2}
    """
    if isinstance(update, dict):
        for key, value in update.items():
            if key in src and isinstance(src[key], dict):
                if isinstance(value, dict):
                    nested_update(src[key], value)
                else:
                    src[key] = value
            else:
                src[key] = value
    return src



# coding: utf-8
# 2021/2/10 @ tongshiwei
from __future__ import absolute_import
from __future__ import print_function

__all__ = ["Configuration", "ConfigurationParser"]

from mxnet import cpu

from longling import path_append
from longling.ML.MxnetHelper.glue import parser
from longling.ML.MxnetHelper.glue.parser import eval_var
from longling.ML.MxnetHelper.toolkit import get_optimizer_cfg
from longling.ML.MxnetHelper.toolkit.select_exp import all_params as _select
from longling.lib.parser import var2exp
from longling.lib.utilog import config_logging
from longling.ML.const import RESULT_JSON, CFG_JSON


class Configuration(parser.Configuration):
    model_name = "model"
    model_dir = model_name

    # 训练参数设置
    begin_epoch = 0
    end_epoch = 100
    batch_size = 32
    save_epoch = 1

    # 优化器设置
    optimizer, optimizer_params = get_optimizer_cfg(name="base")
    # lr_params 中包含 update_params 时，学习率衰减方式运行时确定
    # lr_params = {"update_params": {}}
    lr_params = {}

    # 更新保存参数，一般需要保持一致
    train_select = _select
    save_select = train_select

    # 运行设备
    ctx = cpu()
    train_ctx = None
    eval_ctx = None

    # 工具包参数
    toolbox_params = {}

    # 用户变量
    # 网络超参数
    hyper_params = {}
    # 网络初始化参数
    init_params = {}
    # 损失函数超参数
    loss_params = {}

    # 说明
    caption = ""

    def __init__(self, params_path=None, params_kwargs=None, **kwargs):
        """
        Configuration File, including categories:

        * directory setting
        * optimizer setting
        * training parameters
        * equipment
        * parameters saving setting
        * user parameters

        Parameters
        ----------
        params_path: str
            The path to configuration file which is in json format
        kwargs:
            Parameters to be reset.

        Examples
        --------
        >>> cfg = Configuration(model_name="longling", model_dir="./", ctx="cpu(0)")
        >>> cfg  # doctest: +ELLIPSIS
        logger: <Logger longling (INFO)>
        model_name: longling
        model_dir: ./
        ...
        >>> cfg.var2val("$model_dir")
        './'
        """
        super(Configuration, self).__init__()
        params = self.class_var
        if params_path:
            params.update(self.load_cfg(cfg_path=params_path, **(params_kwargs if params_kwargs else {})))
        params.update(**kwargs)

        self.validation_result_file = None
        self.cfg_path = None

        self.update(**params)

    def _update(self, **kwargs):
        params = kwargs
        params["logger"] = params.pop(
            "logger",
            config_logging(
                logger=params.get("model_name", self.model_name),
                console_log_level="info"
            )
        )

        for key in params:
            if key.endswith("_params") and key + "_update" in params:
                params[key].update(params[key + "_update"])

        self.deep_update(**params)

        _vars = [
            "ctx"
        ]
        for _var in _vars:
            if _var in kwargs:
                try:
                    setattr(self, _var, eval_var(kwargs[_var]))
                except TypeError:
                    pass

        self.validation_result_file = path_append(
            self.model_dir, RESULT_JSON, to_str=True
        )
        self.cfg_path = path_append(
            self.model_dir, CFG_JSON, to_str=True
        )

    def dump(self, cfg_path=None, override=True, file_format=None):
        cfg_path = self.cfg_path if cfg_path is None else cfg_path
        super(Configuration, self).dump(cfg_path, override, file_format=file_format)

    def var2val(self, var):
        return eval(var2exp(
            var,
            env_wrap=lambda x: "self.%s" % x
        ))


class ConfigurationParser(parser.ConfigurationParser):
    pass



# coding: utf-8
# 2020/8/18 @ tongshiwei

import logging
import numpy as np
import scipy.stats as st
from mxnet.lr_scheduler import LRScheduler, FactorScheduler, PolyScheduler, MultiFactorScheduler, CosineScheduler

from longling.ML.DL import get_max_update, get_lr_factor, get_final_lr, get_step

__all__ = ["get_lr_scheduler", "plot_schedule", "get_max_update"]


class _LRScheduler(LRScheduler):
    def __repr__(self):
        return str(self.state_dict)

    def __call__(self, *args, **kwargs):
        raise NotImplementedError

    @property
    def state_dict(self):
        raise NotImplementedError

    @classmethod
    def init(cls, base_lr, batches_per_epoch=None, update_epoch=None, epoch_update_freq=1, warmup_epoch=0,
             discount=None, *args, **kwargs):

        if warmup_epoch:
            kwargs.update({"warmup_steps": warmup_epoch * batches_per_epoch})

        if discount is not None:
            final_lr, discount = get_final_lr(base_lr, discount=discount)
            kwargs.update({"final_lr": final_lr})

        return kwargs


class _FactorScheduler(FactorScheduler, _LRScheduler):
    """update at every *n* steps (batches)"""

    def __repr__(self):
        return str(self.state_dict)

    @property
    def state_dict(self):
        return {
            "scheduler": "FactorScheduler",
            "base_lr": self.base_lr,
            "step": self.step,
            "stop_factor_lr": self.stop_factor_lr,
            "warmup_mode": self.warmup_mode,
            "warmup_begin_lr": self.warmup_begin_lr,
            "warmup_steps": self.warmup_steps,
        }

    @classmethod
    def init(cls, base_lr, batches_per_epoch=None, update_epoch=None, epoch_update_freq=1, warmup_epoch=0,
             discount=None, **kwargs):
        kwargs.update(super(_FactorScheduler, cls).init(
            base_lr,
            warmup_epoch=warmup_epoch,
            batches_per_epoch=batches_per_epoch,
            discount=discount
        ))

        if batches_per_epoch is not None and update_epoch is not None:
            step, factor, stop_factor_lr = get_lr_factor(
                base_lr,
                step=get_step(batches_per_epoch, epoch_update_freq),
                max_update=get_max_update(update_epoch, batches_per_epoch, kwargs.get("warmup_steps", 0)),
                final_lr=kwargs.get("final_lr", kwargs.get("stop_factor_lr", None)),
                discount=discount
            )
            kwargs.update({"step": step, "factor": factor, "stop_factor_lr": stop_factor_lr})

        elif "max_update" in kwargs:
            max_update = kwargs.pop("max_update")
            step, factor, stop_factor_lr = get_lr_factor(
                base_lr,
                step=kwargs["step"],
                max_update=max_update,
                final_lr=kwargs.get("final_lr", kwargs.get("stop_factor_lr", None)),
                discount=discount
            )

            kwargs.update({"step": step, "factor": factor, "stop_factor_lr": stop_factor_lr})

        if "final_lr" in kwargs:
            kwargs.pop("final_lr")

        return cls(base_lr=base_lr, **kwargs)


class _PolyScheduler(PolyScheduler, _LRScheduler):
    """update at each step (batch)"""

    @property
    def state_dict(self):
        return {
            "scheduler": "PolyScheduler",
            "base_lr": self.base_lr,
            "step": 1,
            "max_steps": self.max_steps,
            "final_lr": self.final_lr,
            "warmup_mode": self.warmup_mode,
            "warmup_begin_lr": self.warmup_begin_lr,
            "warmup_steps": self.warmup_steps,
        }

    @classmethod
    def init(cls, base_lr, batches_per_epoch=None, update_epoch=None, epoch_update_freq=0, warmup_epoch=0,
             discount=None, *args, **kwargs):
        kwargs = super(_PolyScheduler, cls).init(
            base_lr,
            warmup_epoch=warmup_epoch,
            batches_per_epoch=batches_per_epoch,
            discount=discount
        )

        if batches_per_epoch is not None and update_epoch is not None:
            max_update = get_max_update(update_epoch, batches_per_epoch, kwargs.get("warmup_steps", 0))
            kwargs.update({"max_update": max_update})

        return cls(base_lr=base_lr, **kwargs)


class _CosineScheduler(CosineScheduler, _LRScheduler):
    """update at each step (batch)"""

    @property
    def state_dict(self):
        return {
            "scheduler": "CosineScheduler",
            "base_lr": self.base_lr,
            "step": 1,
            "max_steps": self.max_steps,
            "final_lr": self.final_lr,
            "warmup_mode": self.warmup_mode,
            "warmup_begin_lr": self.warmup_begin_lr,
            "warmup_steps": self.warmup_steps,
        }

    @classmethod
    def init(cls, base_lr, batches_per_epoch=None, update_epoch=None, epoch_update_freq=0, warmup_epoch=0,
             discount=None, *args, **kwargs):
        batches_per_epoch = epoch_update_freq if batches_per_epoch is None else batches_per_epoch

        kwargs.update(super(_CosineScheduler, cls).init(
            base_lr,
            warmup_epoch=warmup_epoch,
            batches_per_epoch=batches_per_epoch,
            discount=discount
        ))

        if batches_per_epoch is not None and update_epoch is not None:
            max_update = get_max_update(batches_per_epoch, update_epoch, kwargs.get("warmup_steps", 0))
            kwargs.update({"max_update": max_update})

        return cls(base_lr=base_lr, **kwargs)


class _MultiFactorScheduler(MultiFactorScheduler, _LRScheduler):
    """Reduce the learning rate by given a list of steps"""

    @property
    def state_dict(self):
        return {
            "scheduler": "MultiFactorScheduler",
            "base_lr": self.base_lr,
            "steps": self.step,
            "warmup_mode": self.warmup_mode,
            "warmup_begin_lr": self.warmup_begin_lr,
            "warmup_steps": self.warmup_steps,
        }

    @classmethod
    def init(cls, base_lr, batches_per_epoch=None, update_epoch=None, epoch_update_freq=1, warmup_epoch=0,
             discount=None, *args, **kwargs):

        kwargs.update(super(_MultiFactorScheduler, cls).init(
            base_lr,
            warmup_epoch=warmup_epoch,
            batches_per_epoch=batches_per_epoch,
            discount=discount
        ))

        if batches_per_epoch is not None and update_epoch is not None:
            assert isinstance(update_epoch, list)

            if kwargs.get("acc_step", False):
                kwargs.pop("acc_step")
                acc_update_epoch = [update_epoch[0]]
                for i, _step in enumerate(update_epoch[1:]):
                    acc_update_epoch.append(acc_update_epoch[i] + _step)
                update_epoch = acc_update_epoch

            _, factor, stop_factor_lr = get_lr_factor(
                base_lr,
                step=len(update_epoch) - 1,
                max_update=update_epoch[-2],
                final_lr=kwargs.get("final_lr", kwargs.get("stop_factor_lr", None)),
                discount=discount
            )
            kwargs.update({"factor": factor})

            step = [_update_epoch * batches_per_epoch for _update_epoch in update_epoch[:-1]]
            kwargs.update({"step": step})

        elif kwargs.get("acc_step", False):
            kwargs.pop("acc_step")
            step = kwargs.pop("step")
            acc_step = [step[0]]
            for i, _step in enumerate(step[1:]):
                acc_step.append(acc_step[i] + _step)
            kwargs.update({"step": acc_step})

        if "final_lr" in kwargs:
            kwargs.pop("final_lr")
        return cls(base_lr=base_lr, **kwargs)


class LinearScheduler(LRScheduler):
    """update at every *n* steps (batches)"""

    def __init__(self, max_update, step=1, base_lr=0.01, final_lr=0,
                 warmup_steps=0, warmup_begin_lr=0, warmup_mode='linear'):
        super(LinearScheduler, self).__init__(
            base_lr, warmup_steps=warmup_steps, warmup_begin_lr=warmup_begin_lr, warmup_mode=warmup_mode
        )
        assert isinstance(max_update, int)
        if max_update < 1:
            raise ValueError("maximum number of updates must be strictly positive")
        self.max_update = max_update
        self.final_lr = final_lr
        self.count = 0
        self.step = step
        self.max_steps = self.max_update // self.step
        self.increase = (self.final_lr - self.base_lr) / ((self.max_steps - 1) if self.max_steps > 1 else 1)

    def __call__(self, num_update):
        if num_update < self.warmup_steps:
            return self.get_warmup_lr(num_update)
        while num_update > self.count + self.step:
            self.count += self.step
            self.base_lr += self.increase
        if self.base_lr < self.final_lr:
            return self.final_lr
        return self.base_lr


class _LinearScheduler(LinearScheduler, _LRScheduler):
    @property
    def state_dict(self):
        return {
            "scheduler": "LinearScheduler",
            "base_lr": self.base_lr,
            "max_steps": self.max_steps,
            "warmup_mode": self.warmup_mode,
            "warmup_begin_lr": self.warmup_begin_lr,
            "warmup_steps": self.warmup_steps,
        }

    @classmethod
    def init(cls, base_lr, batches_per_epoch=None, update_epoch=None, epoch_update_freq=1, warmup_epoch=0,
             discount=None, *args, **kwargs):
        batches_per_epoch = epoch_update_freq if batches_per_epoch is None else batches_per_epoch

        kwargs.update(super(_LinearScheduler, cls).init(
            base_lr,
            warmup_epoch=warmup_epoch,
            batches_per_epoch=batches_per_epoch,
            discount=discount
        ))
        if batches_per_epoch is not None and update_epoch is not None:
            kwargs.update({"step": get_step(batches_per_epoch, epoch_update_freq)})
            kwargs.update(
                {"max_update": get_max_update(update_epoch, batches_per_epoch, kwargs.get("warmup_steps", 0))}
            )
        return cls(base_lr=base_lr, **kwargs)


class NormScheduler(LRScheduler):
    """update at every *n* steps (batches)"""

    def __init__(self,
                 max_update, step=1,
                 base_lr=0.01, final_lr=1e-8,
                 lr_loc=None, lr_scale=None,
                 min_lr=None, max_lr=None,
                 epsilon=1e-10,
                 warmup_steps=0, warmup_begin_lr=0, warmup_mode='linear'):

        if lr_loc is not None:
            self.lr_loc = lr_loc
            self.lr_scale = lr_scale if lr_scale is not None else 2 * lr_loc
            base_lr = self.lr_loc + 2 * self.lr_scale
        else:
            self.lr_loc = base_lr / 2 if lr_loc is None else lr_loc
            self.lr_scale = lr_scale if lr_scale is not None else (base_lr - self.lr_loc) / 2
        self.min_lr = max(final_lr, min_lr if min_lr is not None else (-2 * self.lr_scale + self.lr_loc))
        self.max_lr = min(max_lr if max_lr is not None else (2 * self.lr_scale + self.lr_loc), base_lr)
        self.dis = st.norm(loc=self.lr_loc, scale=self.lr_scale)
        self._p = 0
        self.epsilon = epsilon
        super(NormScheduler, self).__init__(
            base_lr=self.get_lr(), warmup_steps=warmup_steps,
            warmup_begin_lr=warmup_begin_lr, warmup_mode=warmup_mode,
        )
        self.step = step
        self.max_steps = max_update / self.step
        self.p_increase = 1 / (self.max_steps - 1 if self.max_steps > 1 else 1)
        self.count = 0

    @property
    def p(self):
        return np.clip(self.epsilon, self._p, 1 - self.epsilon)

    def get_lr(self):
        lr = self.dis.isf(self.p)
        if not self.min_lr <= lr <= self.max_lr:
            lr = np.clip(lr, self.min_lr, self.max_lr)
        return lr

    def __call__(self, num_update):
        if num_update < self.warmup_steps:
            return self.get_warmup_lr(num_update)

        # NOTE: use while rather than if  (for continuing training via load_epoch)
        while num_update > self.count + self.step:
            self.count += self.step
            self._p += self.p_increase
            self.base_lr = self.get_lr()
            logging.info("Update[%d]: Change learning rate to %0.5e",
                         num_update, self.base_lr)
        return self.base_lr


class _NormScheduler(NormScheduler, _LRScheduler):
    @property
    def state_dict(self):
        return {
            "scheduler": "NormScheduler",
            "lr_loc": self.lr_loc,
            "step": self.step,
            "max_lr": self.max_lr,
            "min_lr": self.min_lr,
            "lr_scale": self.lr_scale,
            "warmup_mode": self.warmup_mode,
            "warmup_begin_lr": self.warmup_begin_lr,
            "warmup_steps": self.warmup_steps,
        }

    @classmethod
    def init(cls, base_lr=None, batches_per_epoch=None, update_epoch=None, epoch_update_freq=1, warmup_epoch=0,
             discount=None, *args, **kwargs):
        batches_per_epoch = epoch_update_freq if batches_per_epoch is None else batches_per_epoch

        kwargs.update(super(_NormScheduler, cls).init(
            base_lr,
            warmup_epoch=warmup_epoch,
            batches_per_epoch=batches_per_epoch,
            discount=discount,
        ))
        if batches_per_epoch is not None and update_epoch is not None:
            max_update = int(batches_per_epoch * update_epoch) + kwargs.get("warmup_steps", 0)
            kwargs.update({"max_update": max_update, "step": get_step(batches_per_epoch, epoch_update_freq)})
        return cls(base_lr=base_lr, **kwargs)


SCHEDULERS = {
    "linear": _LinearScheduler,
    "factor": _FactorScheduler,
    "poly": _PolyScheduler,
    "multifactor": _MultiFactorScheduler,
    "cosine": _CosineScheduler,
    "norm": _NormScheduler,
}


def get_lr_scheduler(scheduler: (str, LRScheduler) = "cosine", logger=logging, update_params=None, adaptive=True,
                     **kwargs):
    """

    Parameters
    ----------
    scheduler
    logger
    update_params
    adaptive
    kwargs

    Other Parameters
    ----------------
    base_lr: int or float
    learning_rate:
        alias of base_lr
    lr:
        alias of base_lr
    batches_per_epoch: int

    update_epoch: int or list of int
        Changes the learning rate for total n epoch (equal or less than the training epoch).
        Specially, for MultiFactorScheduler, The list of update epochs to schedule a change,
        notice the last one element indicates the total training epoch, i.e., [c1, ..., cn, end_epoch]
    epoch_update_freq: int
        default to 1
        when set to 0, update_freq will be the same as batches_per_epoch (i.e., update once at an epoch)
    max_update: int
        maximum number of updates before the decay reaches final learning rate, which includs warmup steps
    total_step: int
        alias for max_update
    step: int or list of int
        Changes the learning rate for every n updates.
        Specially, for MultiFactorScheduler, The list of steps to schedule a change
    acc_step: bool
        Only for MultiFactorScheduler, whether calculate the accumulative steps based on step
    final_lr: int or float
        final learning rate after all steps
    warmup_epoch: int
    warmup_steps: int
        number of warmup steps used before this scheduler starts decay
    warmup_begin_lr: float
        if using warmup, the learning rate from which it starts warming up
    warmup_mode: str
        warmup can be done in two modes.
        'linear' mode gradually increases lr with each step in equal increments
        'constant' mode keeps lr at warmup_begin_lr for warmup_steps

    Returns
    -------
    scheduler: _LRScheduler or LRScheduler

    Examples
    --------
    >>> get_lr_scheduler(max_update=100, lr=0.01)  # doctest: +NORMALIZE_WHITESPACE
    {'scheduler': 'CosineScheduler', 'base_lr': 0.01, 'step': 1,
    'max_steps': 100, 'final_lr': 0, 'warmup_mode': 'linear', 'warmup_begin_lr': 0, 'warmup_steps': 0}
    >>> get_lr_scheduler("cosine", batches_per_epoch=10, update_epoch=10, lr=0.01)  # doctest: +NORMALIZE_WHITESPACE
    {'scheduler': 'CosineScheduler', 'base_lr': 0.01, 'step': 1,
    'max_steps': 100, 'final_lr': 0, 'warmup_mode': 'linear', 'warmup_begin_lr': 0, 'warmup_steps': 0}
    >>> get_lr_scheduler("CosineScheduler", max_update=100,
    ... base_lr=0.01, adaptive=False)  # doctest: +ELLIPSIS
    <mxnet.lr_scheduler.CosineScheduler...
    >>> get_lr_scheduler("norm", base_lr=1, discount=0.01, max_update=100)   # doctest: +NORMALIZE_WHITESPACE
    {'scheduler': 'NormScheduler', 'lr_loc': 0.5, 'step': 1, 'max_lr': 1.0, 'min_lr': 0.01, 'lr_scale': 0.25,
    'warmup_mode': 'linear', 'warmup_begin_lr': 0, 'warmup_steps': 0}
    >>> get_lr_scheduler("norm", base_lr=1, discount=0.01, lr_loc=0.02,
    ...     batches_per_epoch=100, update_epoch=10, epoch_update_freq=10,
    ... )  # doctest: +NORMALIZE_WHITESPACE
    {'scheduler': 'NormScheduler', 'lr_loc': 0.02, 'step': 10, 'max_lr': 0.1,
    'min_lr': 0.01, 'lr_scale': 0.04, 'warmup_mode': 'linear', 'warmup_begin_lr': 0, 'warmup_steps': 0}
    >>> get_lr_scheduler("linear", learning_rate=0.01, discount=0.1,
    ...     max_update=20, step=2)  # doctest: +NORMALIZE_WHITESPACE
    {'scheduler': 'LinearScheduler', 'base_lr': 0.01, 'max_steps': 10,
    'warmup_mode': 'linear', 'warmup_begin_lr': 0, 'warmup_steps': 0}
    >>> get_lr_scheduler("linear", learning_rate=0.01, discount=0.1,
    ...     update_epoch=2, warmup_epoch=1, batches_per_epoch=10)  # doctest: +NORMALIZE_WHITESPACE
    {'scheduler': 'LinearScheduler', 'base_lr': 0.01, 'max_steps': 3,
    'warmup_mode': 'linear', 'warmup_begin_lr': 0, 'warmup_steps': 10}
    >>> get_lr_scheduler("factor", learning_rate=0.01,
    ...     discount=0.1, max_update=10, step=2)  # doctest: +NORMALIZE_WHITESPACE
    {'scheduler': 'FactorScheduler', 'base_lr': 0.01, 'step': 2, 'stop_factor_lr': 0.001, 'warmup_mode': 'linear',
    'warmup_begin_lr': 0, 'warmup_steps': 0}
    >>> get_lr_scheduler("factor", learning_rate=0.01, discount=0.1,
    ...     update_epoch=2, warmup_epoch=1, batches_per_epoch=10)  # doctest: +NORMALIZE_WHITESPACE
    {'scheduler': 'FactorScheduler', 'base_lr': 0.01, 'step': 10, 'stop_factor_lr': 0.001,
    'warmup_mode': 'linear', 'warmup_begin_lr': 0, 'warmup_steps': 10}
    >>> get_lr_scheduler("poly", learning_rate=0.01, discount=0.1,
    ...     update_epoch=2, warmup_epoch=1, batches_per_epoch=10)  # doctest: +NORMALIZE_WHITESPACE
    {'scheduler': 'PolyScheduler', 'base_lr': 0.01, 'step': 1, 'max_steps': 20,
    'final_lr': 0.001, 'warmup_mode': 'linear', 'warmup_begin_lr': 0, 'warmup_steps': 10}
    >>> get_lr_scheduler("multifactor", learning_rate=0.01, discount=0.1,
    ...     step=[20, 30, 40])  # doctest: +NORMALIZE_WHITESPACE
    {'scheduler': 'MultiFactorScheduler', 'base_lr': 0.01, 'steps': [20, 30, 40],
    'warmup_mode': 'linear', 'warmup_begin_lr': 0, 'warmup_steps': 0}
    >>> get_lr_scheduler("multifactor", learning_rate=0.01, discount=0.1,
    ...     step=[20, 10, 10], acc_step=True)  # doctest: +NORMALIZE_WHITESPACE
    {'scheduler': 'MultiFactorScheduler', 'base_lr': 0.01, 'steps': [20, 30, 40],
    'warmup_mode': 'linear', 'warmup_begin_lr': 0, 'warmup_steps': 0}
    >>> get_lr_scheduler("multifactor", learning_rate=0.01, discount=0.1,
    ...     update_epoch=[2, 3, 4], warmup_epoch=1, batches_per_epoch=10)  # doctest: +NORMALIZE_WHITESPACE
    {'scheduler': 'MultiFactorScheduler', 'base_lr': 0.01, 'steps': [20, 30],
    'warmup_mode': 'linear', 'warmup_begin_lr': 0, 'warmup_steps': 10}
    >>> get_lr_scheduler("multifactor", learning_rate=0.01, discount=0.1, update_epoch=[2, 1, 1],
    ...     warmup_epoch=1, acc_step=True, batches_per_epoch=10)  # doctest: +NORMALIZE_WHITESPACE
    {'scheduler': 'MultiFactorScheduler', 'base_lr': 0.01, 'steps': [20, 30],
    'warmup_mode': 'linear', 'warmup_begin_lr': 0, 'warmup_steps': 10}
    >>> get_lr_scheduler("cosine", learning_rate=0.01, discount=0.1,
    ...     update_epoch=2, warmup_epoch=1, batches_per_epoch=10)  # doctest: +NORMALIZE_WHITESPACE
    {'scheduler': 'CosineScheduler', 'base_lr': 0.01, 'step': 1, 'max_steps': 20, 'final_lr': 0.001,
    'warmup_mode': 'linear', 'warmup_begin_lr': 0, 'warmup_steps': 10}
    >>> get_lr_scheduler("norm", learning_rate=0.01, discount=0.1,
    ...     update_epoch=2, warmup_epoch=1, batches_per_epoch=10)   # doctest: +NORMALIZE_WHITESPACE
    {'scheduler': 'NormScheduler', 'lr_loc': 0.005, 'step': 10, 'max_lr': 0.01, 'min_lr': 0.001,
    'lr_scale': 0.0025, 'warmup_mode': 'linear', 'warmup_begin_lr': 0, 'warmup_steps': 10}
    """
    if adaptive is True:
        for key in {"learning_rate", "lr"}:
            if key in kwargs:
                kwargs["base_lr"] = kwargs.pop(key)

        if isinstance(scheduler, str):
            scheduler = SCHEDULERS[scheduler].init(**kwargs)
    else:
        if isinstance(scheduler, str):
            scheduler = eval(scheduler)(**kwargs)

    assert isinstance(scheduler, LRScheduler)

    logger.info(scheduler)

    return scheduler


def plot_schedule(schedule_fn, iterations=1500, show=True):
    """

    Parameters
    ----------
    schedule_fn
    iterations
    show

    Returns
    -------
    learning rates: list

    Examples
    --------
    >>> scheduler = get_lr_scheduler("linear", base_lr=1, final_lr=0,
    ...     update_epoch=2, warmup_epoch=1, batches_per_epoch=10)
    >>> plot_schedule(scheduler, iterations=20, show=False)   # doctest: +NORMALIZE_WHITESPACE
    [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5]
    >>> scheduler = get_lr_scheduler("norm", base_lr=1, discount=0.01,
    ...     batches_per_epoch=2, update_epoch=9, warmup_epoch=1)
    >>> plot_schedule(scheduler, iterations=20, show=False)   # doctest: +NORMALIZE_WHITESPACE
    [0.5, 1.0, 0.8051600872118375, 0.8051600872118375, 0.6911774184465967, 0.6911774184465967, 0.6076818248238643,
    0.6076818248238643, 0.5349275747204655, 0.5349275747204655, 0.46507242527953446, 0.46507242527953446,
    0.39231817517613554, 0.39231817517613554, 0.30882258155340314, 0.30882258155340314, 0.19483991278816226,
    0.19483991278816226, 0.01, 0.01]
    """
    import matplotlib.pyplot as plt

    # Iteration count starting at 1
    iterations = [i + 1 for i in range(iterations)]
    lrs = [schedule_fn(i) for i in iterations]

    plt.scatter(iterations, lrs)
    plt.xlabel("Iteration")
    plt.ylabel("Learning Rate")
    if show:  # pragma: no cover
        plt.show()
    return lrs


if __name__ == '__main__':
    logging.getLogger().setLevel(logging.INFO)
    _scheduler = get_lr_scheduler(
        "norm", base_lr=1, discount=0.01, lr_loc=0.02,
        batches_per_epoch=100, update_epoch=10, epoch_update_freq=10,
    )
    plot_schedule(_scheduler)



# coding: utf-8
# 2020/4/19 @ tongshiwei
import itertools
import math
import random

___all__ = ["ema", "choice"]


def ema(x, mu=None, alpha=0.3):
    """
    Exponential moving average: smoothing to give progressively lower
    weights to older values.

    Code from tqdm

    Parameters
    ----------
    x  : float
        New value to include in EMA.
    mu  : float, optional
        Previous EMA value.
    alpha  : float, optional
        Smoothing factor in range [0, 1], [default: 0.3].
        Increase to give more weight to recent values.
        Ranges from 0 (yields mu) to 1 (yields x).

    Examples
    -------
    >>> ema(1)
    1
    >>> ema(1, 10)
    7.3
    """
    return x if mu is None else (alpha * x) + (1 - alpha) * mu


def smooth(array: list, smooth_weight: float) -> list:
    """
    smooth the list

    Examples
    --------
    >>> smooth([0, 1, 2, 3, 4], 0)
    [0, 1, 2, 3, 4]
    >>> "%.1f" % smooth([0, 1] * 10000, 0.999)[-1]
    '0.5'

    """
    smoothed = [array[0]]
    for scalar in array[1:]:
        smoothed.append(ema(scalar, smoothed[-1], 1 - smooth_weight))
    return smoothed


def choice(obj) -> int:
    """
    weighted choice

    Examples
    --------
    >>> random.seed(10)
    >>> choice([0.1, 0.2, 0.4, 0.3])
    2
    >>> choice({1: 0.1, 2: 0.9})
    2
    >>> choice({1: 1 / 3, 2: 0.66666})
    2
    """
    r = random.random()

    if isinstance(obj, list):
        assert obj
        if round(sum(obj), 2) != 1:
            raise ValueError("the sum of obj should be 1, now is %s" % sum(obj))
        for i, value in enumerate(itertools.accumulate(obj)):
            if r < value:
                return i
        else:  # pragma: no cover
            return obj[-1]
    elif isinstance(obj, dict):
        assert obj
        keys = obj.keys()
        values = itertools.accumulate(obj.values())
        if round(sum(obj.values()), 2) != 1:
            raise ValueError("the sum of obj should be 1, now is %s" % sum(obj.values()))
        for key, value in zip(keys, values):
            if r < value:
                return key
        else:  # pragma: no cover
            return key
    else:
        raise TypeError("obj should be either list or dict, now is %s" % type(obj))


def embedding_dim(embedding_size):
    """

    Parameters
    ----------
    embedding_size

    Returns
    -------

    Examples
    --------
    >>> embedding_dim(10e4)
    96
    """
    _constant = 8.33
    n = _constant * math.log(embedding_size)
    return int(math.ceil(n))



# coding: utf-8
# create by tongshiwei on 2019-9-1

__all__ = ["get_net", "get_loss"]

# todo: define your network symbol and back propagation loss function

from mxnet import gluon

from longling.ML.MxnetHelper.toolkit import loss_dict2tmt_mx_loss


def get_net(*args, **kwargs):
    return NetName(*args, **kwargs)


def get_loss(**kwargs):
    return loss_dict2tmt_mx_loss({
        "L2Loss": gluon.loss.L2Loss(**kwargs)
    })


class NetName(gluon.HybridBlock):
    def __init__(self, prefix=None, params=None):
        super(NetName, self).__init__(prefix, params)

        with self.name_scope():
            pass

    def hybrid_forward(self, F, x, *args, **kwargs):
        pass



# coding: utf-8
# create by tongshiwei on 2017/12/2
# from __future__ import absolute_import

# py2 aborted
# def unistr(x):
#     return x

# py2 aborted
# def tostr(x):
#     return str(x)

# string_types = basestring,
# integer_types = (int, long)
# unistr = lambda x: x.decode('utf-8')
# tostr = lambda x: unicode(x)



# coding: utf-8
# create by tongshiwei on 2019-8-27

import functools

import fire

from longling import Architecture as arch_cli
from longling.Architecture.cli import cli as arch
from longling.Architecture.install_file import nni as install_nni
from longling.ML import embedding_dim
from longling.ML.toolkit.analyser.cli import select_max, arg_select_max, select_min, arg_select_min
from longling.ML.toolkit.analyser.to_board import to_board
from longling.ML.toolkit.dataset import train_valid_test, train_test, kfold
from longling.ML.toolkit.hyper_search.nni import show_top_k, show
from longling.lib.loading import csv2jsonl, jsonl2csv
from longling.lib.stream import encode
from longling.spider import download_file as download
from longling.toolbox import toc

tarch = functools.partial(arch, author="Shiwei Tong")


def cli():  # pragma: no cover
    fire.Fire(
        {
            "train_valid_test": train_valid_test,
            "train_test": train_test,
            "kfold": kfold,
            "toc": toc,
            "arch": arch,
            "tarch": tarch,
            "arch-cli": arch_cli,
            "max": select_max,
            "amax": arg_select_max,
            "min": select_min,
            "amin": arg_select_min,
            "csv2jsonl": csv2jsonl,
            "jsonl2csv": jsonl2csv,
            "encode": encode,
            "download": download,
            "nni": {
                "k_best": show_top_k,
                "show": show,
            },
            "install": {
                "nni": install_nni,
            },
            "to_board": to_board,
            "ml": {
                "embedding_dim": embedding_dim,
            }
        }
    )


if __name__ == '__main__':
    cli()



# coding: utf-8
# 2020/5/11 @ tongshiwei


def key_parser(key):
    """
    Examples
    --------
    >>> key_parser("macro avg:f1")
    ['macro avg', 'f1']
    >>> key_parser("accuracy")
    'accuracy'
    >>> key_parser("iteration:accuracy")
    ['iteration', 'accuracy']
    """
    if ":" in key:
        # prf:0:f1
        return key.split(":")
    return key


def get_by_key(data, parsed_key):
    _key = key_parser(parsed_key)
    _data = data
    if isinstance(_key, list):
        for k in _key:
            _data = _data[k]
    else:
        _data = _data[_key]
    return _data



# coding: utf-8
# 2020/4/13 @ tongshiwei

import logging
from longling.lib.candylib import as_list
from collections import OrderedDict
from sklearn.metrics import (classification_report as cr,
                             roc_auc_score, average_precision_score, accuracy_score
                             )

from sklearn.utils import check_consistent_length
from sklearn.utils.multiclass import unique_labels
import functools
from longling.ML.metrics.utils import POrderedDict

__all__ = ["classification_report"]


def multiclass2multilabel(y: list):
    """
    Parameters
    ----------
    y: 1d label indicator array

    Examples
    --------
    >>> multiclass2multilabel([0, 1, 5])
    matrix([[1., 0., 0.],
            [0., 1., 0.],
            [0., 0., 1.]])
    """
    from sklearn.preprocessing import OneHotEncoder
    ret = []
    for _y in y:
        ret.append([_y])
    return OneHotEncoder().fit_transform(ret).todense()


def classification_report(y_true, y_pred=None, y_score=None, labels=None, metrics=None, sample_weight=None,
                          average_options=None, multiclass_to_multilabel=False, logger=logging, **kwargs):
    """
    Currently support binary and multiclasss classification.

    Parameters
    ----------
    y_true : list, 1d array-like, or label indicator array / sparse matrix
        Ground truth (correct) target values.

    y_pred : list or None, 1d array-like, or label indicator array / sparse matrix
        Estimated targets as returned by a classifier.

    y_score : array or None, shape = [n_samples] or [n_samples, n_classes]
        Target scores, can either be probability estimates of the positive
        class, confidence values, or non-thresholded measure of decisions
        (as returned by "decision_function" on some classifiers). For binary
        y_true, y_score is supposed to be the score of the class with greater
        label.

    labels : array, shape = [n_labels]
        Optional list of label indices to include in the report.

    metrics: list of str,
        Support: precision, recall, f1, support, accuracy, auc, aupoc.

    sample_weight : array-like of shape = [n_samples], optional
        Sample weights.

    average_options: str or list
        default to macro, choices (one or many): "micro", "macro", "samples", "weighted"

    multiclass_to_multilabel: bool

    logger

    Returns
    -------

    Examples
    --------
    >>> import numpy as np
    >>> # binary classification
    >>> y_true = np.array([0, 0, 1, 1, 0])
    >>> y_pred = np.array([0, 1, 0, 1, 0])
    >>> classification_report(y_true, y_pred)
               precision    recall        f1  support
    0           0.666667  0.666667  0.666667        3
    1           0.500000  0.500000  0.500000        2
    macro_avg   0.583333  0.583333  0.583333        5
    accuracy: 0.600000
    >>> y_true = np.array([0, 0, 1, 1])
    >>> y_score = np.array([0.1, 0.4, 0.35, 0.8])
    >>> classification_report(y_true, y_score=y_score)    # doctest: +NORMALIZE_WHITESPACE
    macro_auc: 0.750000	macro_aupoc: 0.833333
    >>> y_true = np.array([0, 0, 1, 1])
    >>> y_pred = [0, 0, 0, 1]
    >>> y_score = np.array([0.1, 0.4, 0.35, 0.8])
    >>> classification_report(y_true, y_pred, y_score=y_score)    # doctest: +NORMALIZE_WHITESPACE
               precision  recall        f1  support
    0           0.666667    1.00  0.800000        2
    1           1.000000    0.50  0.666667        2
    macro_avg   0.833333    0.75  0.733333        4
    accuracy: 0.750000	macro_auc: 0.750000	macro_aupoc: 0.833333
    >>> # multiclass classification
    >>> y_true = [0, 1, 2, 2, 2]
    >>> y_pred = [0, 0, 2, 2, 1]
    >>> classification_report(y_true, y_pred)
               precision    recall        f1  support
    0                0.5  1.000000  0.666667        1
    1                0.0  0.000000  0.000000        1
    2                1.0  0.666667  0.800000        3
    macro_avg        0.5  0.555556  0.488889        5
    accuracy: 0.600000
    >>> # multiclass in multilabel
    >>> y_true = np.array([0, 0, 1, 1, 2, 1])
    >>> y_pred = np.array([2, 1, 0, 2, 1, 0])
    >>> y_score = np.array([
    ...    [0.15, 0.4, 0.45],
    ...    [0.1, 0.9, 0.0],
    ...    [0.33333, 0.333333, 0.333333],
    ...    [0.15, 0.4, 0.45],
    ...    [0.1, 0.9, 0.0],
    ...    [0.33333, 0.333333, 0.333333]
    ... ])
    >>> classification_report(
    ...    y_true, y_pred, y_score,
    ...    multiclass_to_multilabel=True,
    ...    metrics=["aupoc"]
    ... )
                  aupoc
    0          0.291667
    1          0.416667
    2          0.166667
    macro_avg  0.291667
    >>> classification_report(
    ...     y_true, y_pred, y_score,
    ...    multiclass_to_multilabel=True,
    ...    metrics=["auc", "aupoc"]
    ... )
                    auc     aupoc
    0          0.250000  0.291667
    1          0.055556  0.416667
    2          0.100000  0.166667
    macro_avg  0.135185  0.291667
    macro_auc: 0.194444
    >>> y_true = np.array([0, 1, 1, 1, 2, 1])
    >>> y_pred = np.array([2, 1, 0, 2, 1, 0])
    >>> y_score = np.array([
    ...    [0.45, 0.4, 0.15],
    ...    [0.1, 0.9, 0.0],
    ...    [0.33333, 0.333333, 0.333333],
    ...    [0.15, 0.4, 0.45],
    ...    [0.1, 0.9, 0.0],
    ...    [0.33333, 0.333333, 0.333333]
    ... ])
    >>> classification_report(
    ...    y_true, y_pred,
    ...    y_score,
    ...    multiclass_to_multilabel=True,
    ... )    # doctest: +NORMALIZE_WHITESPACE
               precision    recall        f1   auc     aupoc  support
    0           0.000000  0.000000  0.000000  1.00  1.000000        1
    1           0.500000  0.250000  0.333333  0.25  0.583333        4
    2           0.000000  0.000000  0.000000  0.10  0.166667        1
    macro_avg   0.166667  0.083333  0.111111  0.45  0.583333        6
    accuracy: 0.166667	macro_auc: 0.437500
    >>> classification_report(
    ...    y_true, y_pred,
    ...    y_score,
    ...    labels=[0, 1],
    ...    multiclass_to_multilabel=True,
    ... )    # doctest: +NORMALIZE_WHITESPACE
               precision  recall        f1   auc     aupoc  support
    0               0.00   0.000  0.000000  1.00  1.000000        1
    1               0.50   0.250  0.333333  0.25  0.583333        4
    macro_avg       0.25   0.125  0.166667  0.45  0.583333        5
    accuracy: 0.166667	macro_auc: 0.437500
    """
    if y_pred is not None:
        check_consistent_length(y_true, y_pred)
    if y_score is not None:
        check_consistent_length(y_true, y_score)

    assert y_pred is not None or y_score is not None

    average_options = set(as_list(average_options) if average_options else ["macro"])
    average_label_fmt = "{average}_avg"
    average_metric_fmt = "{average}_{metric}"

    if y_pred is not None:
        _unique_labels = unique_labels(y_true, y_pred)
    else:
        _unique_labels = unique_labels(y_true)

    labels = _unique_labels if labels is None else labels
    labels_set = set(labels)

    if not metrics:
        if y_pred is not None:
            metrics = [
                "accuracy", "precision", "recall", "f1",
            ]
        else:
            metrics = []
        if y_score is not None:
            metrics += [
                "auc", "aupoc",
            ]
        if y_pred is not None:
            metrics += ["support"]
    _metrics = set(metrics)

    ret = OrderedDict()

    if _metrics & {"precision", "recall", "f1", "support", "accuracy"}:
        logger.info("evaluate %s" % ",".join(_metrics & {"precision", "recall", "f1", "support", "accuracy"}))
        cr_result = cr(y_true, y_pred, labels=labels, sample_weight=sample_weight, output_dict=True)

        if "accuracy" in cr_result:
            acc = cr_result.pop("accuracy")
        else:
            acc = accuracy_score(y_true, y_pred)

        if "accuracy" in _metrics:
            ret["accuracy"] = acc

        for key, value in cr_result.items():
            ret[key] = {}

            for k in _metrics & {"precision", "recall", "f1", "support", "accuracy"}:
                _k = k if k != "f1" else "f1-score"
                if _k in value:
                    ret[key][k] = value[_k]

        for average in ["micro", "macro", "samples", "weighted"]:
            _label = average_label_fmt.format(average=average)
            __label = " ".join(_label.split("_"))
            _prefix = __label.split(" ")[0]
            if _prefix in average_options:
                ret[_label] = ret.pop(__label)
            elif __label in ret:
                ret.pop(__label)

    if "auc" in _metrics:
        logger.info("evaluate auc")

        assert y_score is not None, "when evaluate auc, y_score is required"

        func = functools.partial(roc_auc_score, y_score=y_score, sample_weight=sample_weight,
                                 **kwargs.get("auc", {"multi_class": 'ovo'}))

        if multiclass_to_multilabel:
            _y_true = multiclass2multilabel(y_true)
            auc_score = func(y_true=_y_true, average=None)
            for _label, score in enumerate(auc_score):
                if _label not in labels_set:
                    continue
                if str(_label) not in ret:
                    ret[str(_label)] = {}
                ret[str(_label)]["auc"] = score

            for average in average_options:
                auc_score = func(y_true=_y_true, average=average)
                _label = average_label_fmt.format(average=average)
                if _label not in ret:
                    ret[_label] = {}

                ret[_label]["auc"] = auc_score

        for average in average_options:
            auc_score = func(y_true=y_true, average=average)
            _label = average_metric_fmt.format(average=average, metric="auc")
            ret[_label] = auc_score

    if "aupoc" in _metrics:
        logger.info("evaluate aupoc")

        func = functools.partial(average_precision_score, y_score=y_score, sample_weight=sample_weight)

        if multiclass_to_multilabel:
            _y_true = multiclass2multilabel(y_true)
            aupoc = func(y_true=_y_true, average=None)
            for _label, score in enumerate(aupoc):
                if _label not in labels_set:
                    continue
                if str(_label) not in ret:
                    ret[str(_label)] = {}
                ret[str(_label)]["aupoc"] = score
            for average in average_options:
                _label = average_label_fmt.format(average=average)
                aupoc = func(y_true=_y_true, average=average)
                if _label not in ret:
                    ret[_label] = {}
                ret[_label]["aupoc"] = aupoc
        if len(_unique_labels) == 2:
            for average in average_options:
                aupoc = func(y_true=y_true)
                _label = average_metric_fmt.format(average=average, metric="aupoc")
                ret[_label] = aupoc

    logger.info("sorting metrics")
    _ret = POrderedDict()
    for key in ret:
        if isinstance(ret[key], dict):
            _ret[key] = OrderedDict()
            for k in metrics:
                if k in ret[key]:
                    _ret[key][k] = ret[key][k]
        else:
            _ret[key] = ret[key]

    return _ret



# coding: utf-8

"""此模块用以进行流处理"""
# todo 完善注释

from __future__ import print_function

import json
from _io import TextIOWrapper
from contextlib import contextmanager
from tqdm import tqdm
import codecs
import os
import sys
from pathlib import PurePath
import fileinput
from typing import BinaryIO, TextIO

__all__ = ['to_io', 'as_io', 'as_out_io', 'rf_open', 'wf_open', 'rf_group_open', 'wf_group_open', 'close_io',
           'flush_print', 'build_dir', 'json_load',
           'as_io_group', 'as_out_io_group',
           'pickle_load', 'AddPrinter', 'AddObject', 'StreamError', 'check_file',
           'PATH_TYPE', 'encode', 'IO_TYPE', "PATH_IO_TYPE", "block_std"
           ]


class StreamError(Exception):
    pass


PATH_TYPE = (str, PurePath)
IO_TYPE = (TextIOWrapper, TextIO, BinaryIO, codecs.StreamReaderWriter, fileinput.FileInput)
PATH_IO_TYPE = (PATH_TYPE, IO_TYPE)


@contextmanager
def block_std():
    """
    Examples
    --------
    >>> print("hello world")
    hello world
    >>> with block_std():
    ...     print("hello world")
    """
    sys.stdout = open(os.devnull, "w")
    sys.stderr = open(os.devnull, "w")
    yield
    sys.stdout = sys.__stdout__
    sys.stderr = sys.__stderr__


def flush_print(*values, **kwargs):
    """刷新打印函数"""
    sep = kwargs.pop('sep', "")
    end = kwargs.pop('end', "")
    print('\r', *values, sep=sep, end=end, flush=True, **kwargs)


def build_dir(path, mode=0o775, parse_dir=True):
    """
    创建目录，从path中解析出目录路径，如果目录不存在，创建目录

    Parameters
    ----------
    path: str
    mode: int
    parse_dir: bool

    """
    if parse_dir:
        dirname = os.path.dirname(path)
    else:
        dirname = path
    if not dirname or os.path.exists(dirname):
        return path
    os.makedirs(dirname, mode)
    return path


def check_file(filepath, size=None):
    """
    检查文件是否存在，size给定时，检查文件大小是否一致

    Parameters
    ----------
    filepath: str
    size: int

    Returns
    -------
    file exist or not: bool

    """
    if os.path.exists(filepath):
        return size == os.path.getsize(filepath) if size is not None else True
    return False


def rf_open(stream_name: (PATH_IO_TYPE, list, None) = None, encoding='utf-8', **kwargs):
    if stream_name is None:
        return sys.stdin
    if isinstance(stream_name, list):
        if encoding is None:
            return fileinput.input(files=stream_name, **kwargs)
        else:
            return fileinput.input(
                files=stream_name,
                openhook=fileinput.hook_encoded(encoding)
            )
    elif isinstance(stream_name, IO_TYPE):
        return stream_name
    elif isinstance(stream_name, PATH_TYPE):
        encoding = None if kwargs.get("mode") == "rb" else encoding
        return open(stream_name, encoding=encoding, **kwargs)
    else:
        raise TypeError("unknown type: %s(%s)" % (stream_name, type(stream_name)))


def rf_group_open(*files, encoding="utf-8", **kwargs):
    return [rf_open(_file, encoding=encoding, **kwargs) for _file in files]


def json_load(fp: (IO_TYPE, PurePath), **kwargs):
    with as_io(fp) as f:
        return json.load(f, **kwargs)


def pickle_load(fp, encoding='utf-8', mode='rb', **kwargs):
    import pickle
    if isinstance(fp, PATH_TYPE):
        fp = open(fp, mode=mode, **kwargs)
        datas = pickle.load(fp, encoding=encoding, **kwargs)
        fp.close()
        return datas
    else:
        return pickle.load(fp, encoding=encoding, **kwargs)


def wf_open(stream_name: (PATH_IO_TYPE, None) = None, mode="w", encoding="utf-8", **kwargs):
    r"""
    Simple wrapper to codecs for writing.

    stream_name为空时 mode - w 返回标准错误输出 stderr;
    否则，返回标准输出 stdout

    stream_name不为空时，返回文件流

    Parameters
    ----------
    stream_name: str, PurePath or None
    mode: str
    encoding: str
        编码方式，默认为 utf-8
    Returns
    -------
    write_stream: StreamReaderWriter
        返回打开的流

    Examples
    --------
    >>> wf = wf_open(mode="stdout")
    >>> print("hello world", file=wf)
    hello world
    """
    if stream_name is None:
        if stream_name is None and mode in {"w", "wb", "a", "ab"}:
            return sys.stdout
        elif mode == "stdout":
            return sys.stdout
        elif mode == "stderr":
            return sys.stderr
        else:
            raise TypeError("Unknown mode for std mode, only `stdout` and `stderr` are supported.")
    elif isinstance(stream_name, PATH_TYPE):
        build_dir(stream_name)
        if mode == "wb":
            return open(stream_name, mode=mode, **kwargs)
        return codecs.open(stream_name, mode=mode, encoding=encoding, **kwargs)
    elif isinstance(stream_name, IO_TYPE):
        return stream_name
    else:
        raise TypeError("unknown type: %s(%s)" % (stream_name, type(stream_name)))


def wf_group_open(*files, encoding="utf-8", **kwargs):
    return [wf_open(_file, encoding=encoding, **kwargs) for _file in files]


def to_io(stream: (TextIOWrapper, TextIO, BinaryIO, PATH_TYPE, list, None) = None, mode='r', encoding='utf-8',
          **kwargs):
    """
    Convert an object as an io stream, could be a path to file or an io stream.

    Examples
    --------

    .. code-block:: python

        to_io("demo.txt")  # equal to open("demo.txt")
        to_io(open("demo.txt"))  # equal to open("demo.txt")
        a = to_io()  # equal to a = sys.stdin
        b = to_io(mode="w)  # equal to a = sys.stdout
    """
    if isinstance(stream, IO_TYPE):
        return stream
    if 'r' in mode:
        return rf_open(stream, encoding=encoding, mode=mode, **kwargs)
    elif 'w' in mode or 'a' in mode or 'stderr' in mode or 'stdout' in mode:
        return wf_open(stream_name=stream, mode=mode, encoding=encoding)
    else:
        raise ValueError("cannot handle the mode %s" % mode)


def to_io_group(*streams, mode='r', encoding='utf-8', **kwargs):
    return [to_io(stream, mode, encoding, **kwargs) for stream in streams]


def close_io(stream):
    """关闭文件流，忽略 sys.stdin, sys.stdout, sys.stderr"""
    if isinstance(stream, list):
        for _stream in stream:
            close_io(_stream)
    elif stream in {sys.stdin, sys.stdout, sys.stderr}:
        pass
    else:
        try:
            stream.close()
        except Exception:
            raise StreamError('io_close: %s' % stream)


@contextmanager
def as_io(src: (TextIOWrapper, TextIO, BinaryIO, PATH_TYPE, list, None) = None, mode='r', encoding='utf-8', **kwargs):
    """
    with wrapper for to_io function, default mode is "r"

    Examples
    --------

    .. code-block:: python

        with as_io("demo.txt") as f:
            for line in f:
                pass

        # equal to
        with open(demo.txt) as src:
            with as_io(src) as f:
                for line in f:
                    pass

        # from several files
        with as_io(["demo1.txt", "demo2.txt"]) as f:
            for line in f:
                pass

        # from sys.stdin
        with as_io() as f:
            for line in f:
                pass
    """
    _io_object = to_io(src, mode, encoding, **kwargs)
    yield _io_object
    close_io(_io_object)


@contextmanager
def as_io_group(*src, mode='r', encoding='utf-8', **kwargs):
    ios = to_io_group(*src, mode=mode, encoding=encoding, **kwargs)
    yield ios
    close_io(ios)


@contextmanager
def as_out_io(tar: (TextIOWrapper, TextIO, BinaryIO, PATH_TYPE, list, None) = None, mode='w', encoding='utf-8',
              **kwargs):
    """
    with wrapper for to_io function, default mode is "w"

    Examples
    --------

    .. code-block:: python

        with as_out_io("demo.txt") as wf:
            print("hello world", file=wf)

        # equal to
        with open(demo.txt) as tar:
            with as_out_io(tar) as f:
                print("hello world", file=wf)

        # to sys.stdout
        with as_out_io() as wf:
            print("hello world", file=wf)

        # to sys.stderr
        with as_out_io(mode="stderr) as wf:
            print("hello world", file=wf)
    """
    with as_io(tar, mode, encoding, **kwargs) as out_io:
        yield out_io


@contextmanager
def as_out_io_group(*src, mode='w', encoding='utf-8', **kwargs):
    out_ios = to_io_group(*src, mode=mode, encoding=encoding, **kwargs)
    yield out_ios
    close_io(out_ios)


def encode(src, src_encoding, tar, tar_encoding):
    """
    Convert a file in source encoding to target encoding

    Parameters
    ----------
    src
    src_encoding
    tar
    tar_encoding

    Returns
    -------

    """
    with rf_open(src, encoding=src_encoding) as f, wf_open(tar, encoding=tar_encoding) as wf:
        for line in tqdm(f, "encoding %s[%s] --> %s[%s]" % (src, src_encoding, tar, tar_encoding)):
            print(line, end='', file=wf)


class AddObject(object):
    def add(self, *args, **kwargs):  # pragma: no cover
        raise NotImplementedError


class AddPrinter(AddObject):
    """
    以add方法添加文件内容的打印器

    Examples
    --------
    >>> import sys
    >>> printer = AddPrinter(sys.stdout, ensure_io=True)
    >>> printer.add("hello world")
    hello world
    """

    def __init__(self, fp, values_wrapper=lambda x: x, to_io_params=None, ensure_io=False, **kwargs):
        super(AddPrinter, self).__init__()
        self.fp = wf_open(fp, **to_io_params if to_io_params is not None else {}) if not ensure_io else fp
        self.value_wrapper = values_wrapper
        self.kwargs = kwargs

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def add(self, value):
        print(self.value_wrapper(value), file=self.fp, **self.kwargs)

    def close(self):
        close_io(self.fp)

    @property
    def name(self):
        return self.fp.name



# coding: utf-8
# 2021/8/5 @ tongshiwei

from mxnet.gluon.nn import Embedding
from mxnet.ndarray import NDArray

__all__ = ["set_embedding_weight"]


def set_embedding_weight(embedding_layer, weight):
    """
    Parameters
    ----------
    embedding_layer: Embedding
    weight: NDArray

    Examples
    --------
    >>> import mxnet as mx
    >>> from mxnet import gluon
    >>> emb1 = gluon.nn.Embedding(2, 3)
    >>> emb1.initialize()
    >>> emb1.weight.set_data(mx.nd.ones((2, 3)))
    >>> emb1.weight.data()
    <BLANKLINE>
    [[1. 1. 1.]
     [1. 1. 1.]]
    <NDArray 2x3 @cpu(0)>
    >>> set_embedding_weight(emb1, mx.nd.zeros((2, 3)))
    >>> emb1.weight.data()
    <BLANKLINE>
    [[0. 0. 0.]
     [0. 0. 0.]]
    <NDArray 2x3 @cpu(0)>
    """
    embedding_layer.weight.set_data(weight)



# coding: utf-8
# create by tongshiwei on 2019-9-4

from __future__ import absolute_import
from __future__ import print_function

import argparse
import os

from longling.lib.parser import path_append
from longling.lib.stream import wf_open
from longling.lib.utilog import config_logging, LogLevel

logger = config_logging(
    logger="glue", level=LogLevel.DEBUG, console_log_level=LogLevel.DEBUG
)


def new_model(model_name, source_dir, directory=None, level="project", skip_existing=False):  # pragma: no cover
    target_dir = os.path.join(
        directory, model_name
    ) if directory is not None else model_name

    target_dir = os.path.abspath(target_dir)
    if level == "project":
        pass
    elif level == "model":
        source_dir = path_append(
            source_dir, "ModelName", to_str=True,
        )
    elif level == "module":
        source_dir = path_append(
            source_dir, "ModelName", "Module", to_str=True,
        )
    else:
        raise ValueError("unknown level: %s" % level)
    if not os.path.exists(source_dir):
        logger.error(
            "template files does not exist, process aborted, "
            "check the path %s" % source_dir
        )
        exit(-1)

    logger.info(source_dir + " -> " + target_dir)
    if os.path.isdir(target_dir):
        logger.error(
            "directory already existed, will not override, generation abort"
        )
        return False
    logger.info("generating file, root path is %s", target_dir)
    big_module_name = "%sModule" % (model_name[0].upper() + model_name[1:])

    def name_replace(name):
        return name.replace(
            "ModelName", model_name
        ).replace("GluonModule", big_module_name)

    for root, dirs, files in os.walk(source_dir):
        if 'data' + os.sep in root or '_build' in root or os.path.join(
                'docs',
                'source'
        ) in root:
            logger.debug("skip %s-%s" % (root, files))
            continue
        for filename in files:
            if '.pyc' in filename:
                logger.debug("skip %s-%s" % (root, filename))
                continue
            dirname = os.path.abspath(target_dir + os.sep + name_replace(
                root.replace(source_dir, "")))
            source_file = os.path.abspath(os.path.join(root, filename))
            target_file = os.path.abspath(
                os.path.join(dirname, name_replace(filename)))
            logger.debug(source_file + ' -> ' + target_file)
            if os.path.exists(target_file) and skip_existing:
                pass
            else:
                with open(source_file, encoding="utf-8") as f, wf_open(
                        target_file) as wf:
                    try:
                        for line in f:
                            print(name_replace(line), end="", file=wf)
                    except UnicodeDecodeError:
                        print(source_file, line)
                        raise UnicodeDecodeError
    return True


def cli(source_dir, model_name=None):  # pragma: no cover
    parser = argparse.ArgumentParser()
    if model_name is not None:
        parser.add_argument(
            "--model_name", default="%s" % model_name,
            help="set the model name, default is %s" % model_name
        )
    else:
        parser.add_argument(
            "--model_name",
            help="set the model name, default is %s" % model_name,
            required=True
        )

    parser.add_argument(
        "--directory", default=None,
        help="set the directory, default is None"
    )
    parser.add_argument(
        "--level", choices={"project", "model", "module"},
        help="the level",
        default="model",
    )
    parser.add_argument(
        "--no-skip-existing",
        help="override the existing files",
        action="store_false"
    )

    args = parser.parse_args()

    if new_model(
            model_name=args.model_name, source_dir=source_dir,
            directory=args.directory, level=args.level, skip_existing=not args.no_skip_existing,
    ):
        logger.info("success")
    else:
        logger.error("error")



# coding: utf-8
# create by tongshiwei on 2018/8/5

from __future__ import absolute_import
from __future__ import print_function

import functools
import os

import longling
from longling.ML.DL import ecli
from longling.lib.parser import path_append

__all__ = ["cli"]

GLUE_DIR = path_append(
    os.path.dirname(longling.__file__),
    "ML.MxnetHelper.glue.ModelName".replace(".", os.sep),
    to_str=True
)
cli = functools.partial(ecli, source_dir=GLUE_DIR)

if __name__ == '__main__':
    cli()



# coding: utf-8
# create by tongshiwei on 2019-9-1
__all__ = ["IterableMIcing", "MonitorPlayer", "ProgressMonitor", "AsyncMonitorPlayer"]

"""
进度监视器，帮助用户知晓当前运行进度，主要适配于机器学习中分 epoch，batch 的情况。

和 tqdm 针对单个迭代对象进行快速适配不同，
progress的目标是能将监视器不同功能部件模块化后再行组装，可以实现description的动态化，
给用户提供更大的便利性。

* MonitorPlayer 定义了如何显示进度和其它过程参数(better than tqdm, where only n is changed and description is fixed)
    * 在 __call__ 方法中定义如何显示
* 继承ProgressMonitor并传入必要参数进行实例化
    * 继承重写ProgressMonitor的__call__函数，用 IterableMIcing 包裹迭代器，这一步可以灵活定义迭代前后的操作
    * 需要在__init__的时候传入一个MonitorPlayer实例
* IterableMIcing 用来组装迭代器、监控器

一个简单的示例如下

.. code-block:: python

    class DemoMonitor(ProgressMonitor):
        def __call__(self, iterator):
            return IterableMIcing(
                iterator,
                self.player, self.player.set_length
            )

    progress_monitor = DemoMonitor(MonitorPlayer())

    for _ in range(5):
        for _ in progress_monitor(range(10000)):
            pass
        print()

cooperate with tqdm

.. code-block:: python

    from tqdm import tqdm

    class DemoTqdmMonitor(ProgressMonitor):
        def __call__(self, iterator, **kwargs):
            return tqdm(iterator, **kwargs)
"""

from typing import Iterable
from longling.lib.stream import flush_print
import threading
import queue


def pass_function(*args, **kwargs):
    pass


class IterableMIcing(Iterable):
    """
    将迭代器包装为监控器可以使用的迭代类：
    * 添加计数器 count, 每迭代一次，count + 1, 迭代结束时，可根据 count 得知数据总长
    * 每次 __iter__ 时会调用 call_in_iter 函数
    * 迭代结束时，会调用 call_after_iter

    Parameters
    ----------
    iterator:
        待迭代数据
    hook_in_iter:
        每次迭代中的回调函数（例如：打印进度等）,接受当前的 count 为输入
    hook_after_iter:
        每轮迭代后的回调函数（所有数据遍历一遍后），接受当前的 length 为输入
    length:
        数据总长（有多少条数据）

    >>> iterator = IterableMIcing(range(100))
    >>> for i in iterator:
    ...     pass
    >>> len(iterator)
    100
    >>> def iter_fn(num):
    ...     for i in range(num):
    ...         yield num
    >>> iterator = IterableMIcing(iter_fn(50))
    >>> for i in iterator:
    ...     pass
    >>> len(iterator)
    50
    """

    def __init__(self,
                 iterator: (Iterable, list, tuple, dict),
                 hook_in_iter=pass_function, hook_after_iter=pass_function,
                 length: (int, None) = None,
                 ):
        self.iterator = iter(iterator)
        self.call_in_iter = hook_in_iter
        self.call_after_iter = hook_after_iter

        self._count = 0
        try:
            if isinstance(length, int):
                self._length = length
            elif length is None:
                self._length = len(iterator)
            else:
                raise TypeError()
        except TypeError:
            self._length = None

    def set_length(self, length):
        self._length = length

    def reset(self, iterator):
        self.iterator = iter(iterator)
        self._count = 0
        self._length = None

    def __len__(self):
        """TypeError will be raised when _length is None."""
        return self._length

    def __next__(self):
        try:
            res = self.iterator.__next__()
            self._count += 1
            self.call_in_iter(self._count)
            return res
        except StopIteration:
            self._length = self._count
            self._count = 0
            self.call_after_iter(self._length)
            raise StopIteration

    def __iter__(self):
        return self


class AsyncMonitorPlayer(object):  # pragma: no cover
    """异步监控器显示器"""

    def __init__(self, cache_size=10000):
        import warnings
        warnings.warn("dev version, do not use")

        self._count = 0
        self._length = None
        self._size = cache_size
        self.thread = None
        self._display_cache = None
        self.reset()

    def display(self):
        while True:
            _count = self._display_cache.get()
            if isinstance(_count, StopIteration):
                return
            self._count = _count
            flush_print("%s|%s" % (self._count, self._length))

    def reset(self):
        self._count = 0
        self._length = None
        if self.thread is not None:
            self._display_cache.put(StopIteration())
            self.thread.join()
        self.thread = threading.Thread(target=self.display, daemon=True)
        self._display_cache = queue.Queue(self._size)
        self.thread.start()

    def __call__(self, _count):
        self._display_cache.put(_count)

    def set_length(self, _length):
        self._length = _length


class MonitorPlayer(object):
    """异步监控器显示器"""

    def __init__(self):
        self._count = 0
        self._length = None

    def reset(self):
        self._count = 0
        self._length = None

    def __call__(self, _count):
        self._count = _count
        flush_print("%s|%s" % (self._count, self._length))

    def set_length(self, _length):
        self._length = _length


class ProgressMonitor(object):
    def __init__(self, player=None):
        self.player = player

    def __call__(self, iterator, *args, **kwargs):
        raise NotImplementedError



# coding: utf-8
# 2019/11/28 @ tongshiwei

import warnings
import functools
import math
from collections import Iterable

try:
    NAN = math.nan
except (AttributeError, ImportError):  # pragma: no cover
    NAN = float('nan')

__all__ = ["ValueMonitor", "EMAValue", "as_tmt_value"]


class BaseTMTValue(object):
    """Basic Template is Module & module is Template (TMT) value"""

    def __init__(self):
        self.pipe = []

    def pipe_put(self, loss_value, *args, **kwargs):
        for _pipe in self.pipe:
            _pipe(loss_value, *args, **kwargs)


class TMTValue(BaseTMTValue):
    def __init__(self, value_obj, transform=lambda x: x):
        super(TMTValue, self).__init__()
        self._value_obj = value_obj
        self._transform = transform

    def __call__(self, *args, **kwargs):
        value = self._value_obj(*args, **kwargs)
        super(TMTValue, self).pipe_put(self._transform(value))
        return value


def as_tmt_value(value_obj, transform=lambda x: x):
    """

    Parameters
    ----------
    value_obj
    transform

    Returns
    -------

    Examples
    --------
    >>> def loss_f(a):
    ...     return a
    >>> loss_f(10)
    10
    >>> tmt_loss_f = as_tmt_value(loss_f)
    >>> tmt_loss_f(10)
    10
    >>> @as_tmt_value
    ... def loss_f2(a):
    ...     return a
    >>> loss_f2(10)
    10
    """
    return TMTValue(value_obj, transform)


# def tmt_value(transform=lambda x: x):
#     def _tmt_value(class_type):
#         class _TMTValue(class_type, BaseTMTValue):
#             @functools.wraps(class_type.__call__)
#             def __call__(self, *args, **kwargs):
#                 value = class_type.__call__(self, *args, **kwargs)
#                 BaseTMTValue.pipe_put(self, self.transform(value))
#                 return value
#
#             @staticmethod
#             def transform(loss):
#                 return transform(loss)
#
#         return _TMTValue
#
#     return _tmt_value


class ValueMonitor(object):
    """
    记录损失函数并显示

    几个重要的函数：
    * update: 定义损失函数的如何累加更新

    Examples
    --------
    >>> vm = ValueMonitor(["m1", "m2"])
    >>> vm.value
    {'m1': nan, 'm2': nan}
    >>> vm.keys()
    dict_keys(['m1', 'm2'])
    >>> vm.values()
    dict_values([nan, nan])
    >>> vm.items()
    dict_items([('m1', nan), ('m2', nan)])
    >>> "m1" in vm
    True
    >>> vm.monitor_off("m1")
    >>> "m1" in vm
    False
    >>> vm.monitor_on("m1")
    >>> "m1" in vm
    True
    >>> def m1_func():
    ...     return 1
    >>> def m2_func():
    ...     return 2
    >>> vm = ValueMonitor({"m1": m1_func, "m2": m2_func}, auto="ignore")
    >>> vm.value
    {'m1': nan, 'm2': nan}
    >>> vm["m1"]
    nan
    """

    def __init__(self, value_function_names: (list, dict), digits=6, auto=None, *args, **kwargs):
        """

        Parameters
        ----------
        value_function_names
        digits
        auto: str, None or bool
        args
        kwargs
        """
        self._values = {name: NAN for name in value_function_names}
        self._funcs = value_function_names if isinstance(value_function_names, dict) else {}
        self._auto = auto
        if self._auto is not False:
            for name, func in self._funcs.items():
                self.link(name, func)
        self._digits = digits

    def __str__(self):
        return str(self._values)

    def __call__(self, *args, **kwargs):
        self.update(*args, **kwargs)

    def __getitem__(self, item):
        return self._values[item]

    def __contains__(self, name):
        return name in self._values

    def monitor_on(self, name, func=None):
        self._values[name] = NAN
        if func is not None:
            self._funcs[name] = func

    def monitor_off(self, name):
        del self._values[name]
        if name in self._funcs:
            del self._funcs[name]

    def get_update_value(self, name, value):
        raise NotImplementedError

    def update(self, name, value):
        self._values[name] = self.round(self.get_update_value(name, value))

    def reset(self):
        for name in self._values:
            self._values[name] = NAN

    def values(self):
        return self._values.values()

    def items(self):
        return self._values.items()

    def keys(self):
        return self._values.keys()

    @property
    def func(self):
        return self._funcs

    @property
    def value(self):
        return self._values

    def round(self, value):
        return round(value, self._digits) if self._digits else value

    def pipe_get(self, name, value):
        if name in self._values:
            self.update(name, value)

    def link(self, name, obj):
        if hasattr(obj, "pipe"):
            obj.pipe.append(functools.partial(self.pipe_get, name))
        else:
            hint = "invalid link operation, for %s has no pipe attribute, suggest to use TMTValue" % obj
            if self._auto is True:
                raise AttributeError(hint)
            elif self._auto is None or self._auto == "warn":
                warnings.warn(hint)


class EMAValue(ValueMonitor):
    """
    Exponential moving average: smoothing to give progressively lower
    weights to older values.

    .. math ::
        losses[name] = (1 - c) \\times previous\\_value + c \\times loss\\_value

    >>> ema = EMAValue(["l2"])
    >>> ema["l2"]
    nan
    >>> ema("l2", 100)
    >>> ema("l2", 1)
    >>> ema["l2"]
    90.1
    >>> list(ema.values())
    [90.1]
    >>> list(ema.keys())
    ['l2']
    >>> list(ema.items())
    [('l2', 90.1)]
    >>> ema.reset()
    >>> ema["l2"]
    nan
    >>> ema = EMAValue(["l1", "l2"])
    >>> ema["l2"], ema["l1"]
    (nan, nan)
    >>> ema.updates({"l1": 1, "l2": 10})
    >>> ema.updates({"l1": 10, "l2": 100})
    >>> ema["l1"]
    1.9
    >>> ema["l2"]
    19.0
    >>> ema = EMAValue(["l1"], smoothing_constant=0.0)
    >>> ema["l1"]
    nan
    >>> ema.updates({"l1": 1})
    >>> ema.updates({"l1": 10})
    >>> ema["l1"]
    1.0
    >>> ema = EMAValue(["l1"], smoothing_constant=1.0)
    >>> ema.updates({"l1": 1})
    >>> ema.updates({"l1": 10})
    >>> ema["l1"]
    10.0
    >>> @as_tmt_value
    ... def mse_loss(a):
    ...     return a ** 2
    >>> ema = EMAValue({"mse": mse_loss})
    >>> ema["mse"]
    nan
    >>> mse_loss(1)
    1
    >>> ema["mse"]
    1
    >>> mse_loss(10)
    100
    >>> ema["mse"]
    10.9
    >>> ema = EMAValue({"mse": mse_loss})
    >>> mse_loss(1)
    1
    >>> ema["mse"]
    1
    >>> ema.monitor_off("mse")
    >>> ema.func
    {}
    >>> mse_loss(10)
    100
    >>> "mse" not in ema
    True
    >>> ema.monitor_on("mse", mse_loss)
    >>> mse_loss(10)
    100
    >>> ema["mse"]
    100
    """

    def __init__(self, value_function_names: (list, dict), smoothing_constant=0.1, *args, **kwargs):
        """
        Parameters
        ----------
        value_function_names
        smoothing_constant: float, optional
            Smoothing factor in range [0, 1], [default: 0.01].
            Increase to give more weight to recent values.
        """
        super(EMAValue, self).__init__(value_function_names, *args, **kwargs)
        self.smoothing_constant = smoothing_constant

    def get_update_value(self, name: str, value: (float, int)):
        """
        Parameters
        ----------
        name: str
            The name of the value to be updated
        value  : int or float
            New value to include in EMA.
        """
        return (
            value if math.isnan(self._values[name])
            else (1 - self.smoothing_constant) * self._values[
                name] + self.smoothing_constant * value
        )

    def updates(self, pairs: (dict, Iterable)):
        pairs = pairs.items() if isinstance(pairs, dict) else pairs
        for name, value in pairs:
            self.update(name, value)



# coding: utf-8
# 2020/8/23 @ tongshiwei


from collections import Iterable
import bisect

__all__ = ["SortedList"]


class SortedList(list):
    """
    A list maintaining the element in an ascending order.

    A custom key function can be supplied to customize the sort order.


    Examples
    --------
    >>> sl = SortedList()
    >>> sl.adds(*[1, 2, 3, 4, 5])
    >>> sl
    [1, 2, 3, 4, 5]
    >>> sl.add(7)
    >>> sl
    [1, 2, 3, 4, 5, 7]
    >>> sl.add(6)
    >>> sl
    [1, 2, 3, 4, 5, 6, 7]
    >>> sl = SortedList([4])
    >>> sl.add(3)
    >>> sl.add(2)
    >>> sl
    [2, 3, 4]
    >>> list(reversed(sl))
    [4, 3, 2]
    >>> sl = SortedList([("harry", 1), ("tom", 0)], key=lambda x: x[1])
    >>> sl
    [('tom', 0), ('harry', 1)]
    >>> sl.add(("jack", -1), key=lambda x: x[1])
    >>> sl
    [('jack', -1), ('tom', 0), ('harry', 1)]
    >>> sl.add(("ada", 2))
    >>> sl
    [('jack', -1), ('tom', 0), ('harry', 1), ('ada', 2)]
    """

    def __init__(self, iterable: Iterable = (), key=None):
        super(SortedList, self).__init__(iterable)
        self._key = key
        self.sort(key=self._key)

    def _get_key(self, elem, key=None):
        if key is not None:
            return key(elem)
        elif self._key is not None:
            return self._key(elem)
        else:
            return elem

    def add(self, elem, key=None):
        if not self:
            self.append(elem)
        elif self._get_key(elem, key) >= self._get_key(self[-1], key):
            self.append(elem)
        elif self._get_key(elem, key) <= self._get_key(self[0], key):
            self.insert(0, elem)
        else:
            idx = bisect.bisect(self, elem)
            self.insert(idx, elem)

    def adds(self, *elem):
        for _elem in elem:
            self.add(_elem)



# coding: utf-8
# create by tongshiwei on 2018/8/5

from __future__ import absolute_import

from .ModelName import ModelName as MetaModel
from .ModelName.ModelName.Module import Module as MetaModule, Configuration as MetaCfg, net_init, fit_f
from .glue import cli

__all__ = ["MetaModel", "MetaModule", "MetaCfg", "net_init", "fit_f"]



# coding: utf-8
# create by tongshiwei on 2019/4/7

import numpy as np
from mxnet import symbol, ndarray

__all__ = ["getF", "copy_net", "array_equal"]


def getF(input_object: (symbol.Symbol, ndarray.NDArray)) -> (symbol, ndarray):
    r"""
    判断输入参数类型

    Parameters
    ----------
    input_object

    Returns
    -------
    The type of input object

    Examples
    --------
    >>> import mxnet as mx
    >>> getF(mx.nd.zeros((3, 3)))   # doctest: +ELLIPSIS
    <module 'mxnet.ndarray'...
    >>> getF(mx.symbol.zeros((3, 3)))   # doctest: +ELLIPSIS
    <module 'mxnet.symbol'...
    """
    if isinstance(input_object, symbol.Symbol):
        return symbol
    elif isinstance(input_object, ndarray.NDArray):
        return ndarray
    else:
        raise TypeError(
            "the type of input should be either Symbol or NDArray, now is %s"
            % type(input_object)
        )


def copy_net(src_net, target_net, select=None):
    """
    复制网络

    Parameters
    ----------
    src_net
    target_net
    select

    Examples
    --------
    >>> import mxnet as mx
    >>> from mxnet import gluon
    >>> emb1 = gluon.nn.Embedding(2, 3)
    >>> emb1.initialize()
    >>> emb1.weight.set_data(mx.nd.ones((2, 3)))
    >>> emb1.weight.data()
    <BLANKLINE>
    [[1. 1. 1.]
     [1. 1. 1.]]
    <NDArray 2x3 @cpu(0)>
    >>> emb2 = gluon.nn.Embedding(2, 3)
    >>> emb2.initialize()
    >>> emb2.weight.set_data(mx.nd.zeros((2, 3)))
    >>> emb2.weight.data()
    <BLANKLINE>
    [[0. 0. 0.]
     [0. 0. 0.]]
    <NDArray 2x3 @cpu(0)>
    >>> copy_net(emb1, emb2)
    >>> emb2.weight.data()
    <BLANKLINE>
    [[1. 1. 1.]
     [1. 1. 1.]]
    <NDArray 2x3 @cpu(0)>
    """
    src_params = src_net.collect_params(select=select)
    target_params = target_net.collect_params(select=select)

    for name, value in src_params._params.items():
        value = value.data()
        target_params._params[
            name.replace(src_params.prefix, target_params.prefix)
        ].set_data(value)


def array_equal(a1, a2):
    """

    Parameters
    ----------
    a1: ndarray.NDArray
    a2: ndarray.NDArray

    Returns
    -------
    b: bool
        Returns True if the arrays are equal.

    Examples
    --------
    >>> import mxnet.ndarray as nd
    >>> a = nd.ones((1, 3))
    >>> a1 = nd.ones((1, 3))
    >>> a2 = nd.zeros((1, 3))
    >>> array_equal(a, a1)
    True
    >>> array_equal(a1, a2)
    False
    """
    return np.array_equal(a1.asnumpy(), a2.asnumpy())
